import base64
import re
from pathlib import Path
from urllib.parse import urlparse

from pydantic import BaseModel, Field

from ..typings import PydanticModelConfig, is_http_url

URL_PATTERN = re.compile(r"https?://[^\s<>\[\]\"'`]+", re.IGNORECASE)
URL_TRAILING_CHARS = ".,;:!?)]}>\"'"
IMAGE_EXTENSIONS = frozenset(
    {
        ".jpg",
        ".jpeg",
        ".png",
        ".gif",
        ".webp",
        ".bmp",
        ".svg",
        ".tif",
        ".tiff",
        ".heic",
        ".heif",
        ".avif",
    }
)


class PromptUrlExtractionResult(BaseModel):
    model_config = PydanticModelConfig.default()
    all_urls: list[str] = Field(default_factory=list)
    image_urls: list[str] = Field(default_factory=list)
    webpage_urls: list[str] = Field(default_factory=list)


def convert_image_to_base64(image_path: Path) -> str | None:
    """Convert an image file to a base64 encoded string."""
    if image_path is None:
        return None
    with open(image_path, "rb") as image_file:
        image_data = image_file.read()
        return base64.b64encode(image_data).decode("utf-8")


def is_image_url(url: str) -> bool:
    parsed = urlparse(url)
    path = parsed.path.lower()
    return any(path.endswith(extension) for extension in IMAGE_EXTENSIONS)


def extract_urls_from_prompt(prompt: str | None) -> PromptUrlExtractionResult:
    if prompt is None or prompt.strip() == "":
        return PromptUrlExtractionResult()

    ordered_urls: list[str] = []
    seen_urls: set[str] = set()

    for raw_url in URL_PATTERN.findall(prompt):
        normalized_url = raw_url.rstrip(URL_TRAILING_CHARS)
        if normalized_url == "":
            continue
        if not is_http_url(normalized_url):
            continue
        if normalized_url in seen_urls:
            continue
        seen_urls.add(normalized_url)
        ordered_urls.append(normalized_url)

    image_urls = [url for url in ordered_urls if is_image_url(url)]
    image_url_set = set(image_urls)
    webpage_urls = [url for url in ordered_urls if url not in image_url_set]

    return PromptUrlExtractionResult(
        all_urls=ordered_urls,
        image_urls=image_urls,
        webpage_urls=webpage_urls,
    )



import uuid
from typing import Annotated, Any, List, Type, Union, get_args, get_origin
from uuid import UUID as _UUID

from bson import (
    Binary,  # require pymongo as dependency, since bson standalone package does not have Binary
)
from bson import ObjectId as _ObjectId
from pydantic import GetCoreSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema


def is_optional(field):
    return get_origin(field) is Union and type(None) in get_args(field)


class ObjectIdPydanticAnnotation:
    @classmethod
    def validate_object_id(cls, v: Any, handler) -> _ObjectId:
        if isinstance(v, _ObjectId):
            return v

        s = handler(v)
        if _ObjectId.is_valid(s):
            return _ObjectId(s)
        else:
            raise ValueError("Invalid ObjectId")

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type, _handler
    ) -> core_schema.CoreSchema:
        assert source_type is _ObjectId or (
            is_optional(source_type) and _ObjectId in get_args(source_type)
        ), (
            "[Db] Internal error, expected source_type is _ObjectId or (is_optional(source_type) and _ObjectId in get_args(source_type))"
        )
        return core_schema.no_info_wrap_validator_function(
            cls.validate_object_id,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema, handler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())


ObjectId = Annotated[_ObjectId, ObjectIdPydanticAnnotation]


class Uuid(_UUID):
    @classmethod
    def new(cls) -> _UUID:
        return uuid.uuid4()

    @classmethod
    def validate_uuid(cls, v: Any) -> _UUID:
        if isinstance(v, _UUID):
            return v
        if isinstance(v, str):
            return _UUID(v)
        if isinstance(v, Binary):
            return v.as_uuid()
        raise ValueError(f"[Db] Unidentifieable UUID data type: {v}")

    # def to_binary(self):
    #     return Binary.from_uuid(self)

    @classmethod
    def from_uuid(cls, payload) -> Binary:
        if isinstance(payload, Binary):
            return payload
        return Binary.from_uuid(payload)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.validate_uuid,
            core_schema.uuid_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                function=cls.from_uuid
            ),
        )


# add json plain serializer to Uuid
# Uuid = Annotated[
#     Uuid, PlainSerializer(lambda x: str(x), return_type=str, when_used='json') # this will override the __get_pydantic_core_schema__ declared above.
# ]

MaybeUuid = Union[Uuid, _UUID, Binary]


def to_binary_uuid(uuid_: Union[MaybeUuid, List[MaybeUuid]]):
    if isinstance(uuid_, list):
        return [to_binary_uuid(u) for u in uuid_]
    if isinstance(uuid_, Binary):
        return uuid_
    return Binary.from_uuid(uuid_)



class Aliyun:
    def __init__(self, *, access_key: str, secret: str):
        self.__access_key = access_key
        self.__secret = secret

    @property
    def access_key(self):
        return self.__access_key

    @property
    def secret(self):
        return self.__secret



import dataclasses
import datetime
from collections import defaultdict, deque
from decimal import Decimal
from enum import Enum
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import Path, PurePath
from re import Pattern
from types import GeneratorType
from typing import Any, Callable, Dict, List, Literal, Optional, Set, Tuple, Type, Union
from uuid import UUID

from pydantic import BaseModel
from pydantic.color import Color
from pydantic.networks import AnyUrl, NameEmail
from pydantic.types import SecretBytes, SecretStr
from pydantic_core import Url as Url


def _model_dump(
    model: BaseModel, mode: Literal["json", "python"] = "json", **kwargs: Any
) -> Any:
    return model.model_dump(mode=mode, **kwargs)


IncEx = Union[Set[int], Set[str], Dict[int, Any], Dict[str, Any]]

PYDANTIC_V2 = True


# Taken from Pydantic v1 as is
def isoformat(o: Union[datetime.date, datetime.time]) -> str:
    return o.isoformat()


# Taken from Pydantic v1 as is
# TODO: pv2 should this return strings instead?
def decimal_encoder(dec_value: Decimal) -> Union[int, float]:
    """
    Encodes a Decimal as int of there's no exponent, otherwise float

    This is useful when we use ConstrainedDecimal to represent Numeric(x,0)
    where a integer (but not int typed) is used. Encoding this as a float
    results in failed round-tripping between encode and parse.
    Our Id type is a prime example of this.

    >>> decimal_encoder(Decimal("1.0"))
    1.0

    >>> decimal_encoder(Decimal("1"))
    1
    """
    if dec_value.as_tuple().exponent >= 0:  # type: ignore[operator]
        return int(dec_value)
    else:
        return float(dec_value)


ENCODERS_BY_TYPE: Dict[Type[Any], Callable[[Any], Any]] = {
    bytes: lambda o: o.decode(),
    Color: str,
    datetime.date: isoformat,
    datetime.datetime: isoformat,
    datetime.time: isoformat,
    datetime.timedelta: lambda td: td.total_seconds(),
    Decimal: decimal_encoder,
    Enum: lambda o: o.value,
    frozenset: list,
    deque: list,
    GeneratorType: list,
    IPv4Address: str,
    IPv4Interface: str,
    IPv4Network: str,
    IPv6Address: str,
    IPv6Interface: str,
    IPv6Network: str,
    NameEmail: str,
    Path: str,
    Pattern: lambda o: o.pattern,
    SecretBytes: str,
    SecretStr: str,
    set: list,
    UUID: str,
    Url: str,
    AnyUrl: str,
}


def generate_encoders_by_class_tuples(
    type_encoder_map: Dict[Any, Callable[[Any], Any]],
) -> Dict[Callable[[Any], Any], Tuple[Any, ...]]:
    encoders_by_class_tuples: Dict[Callable[[Any], Any], Tuple[Any, ...]] = defaultdict(
        tuple
    )
    for type_, encoder in type_encoder_map.items():
        encoders_by_class_tuples[encoder] += (type_,)
    return encoders_by_class_tuples


encoders_by_class_tuples = generate_encoders_by_class_tuples(ENCODERS_BY_TYPE)


def jsonable_encoder(
    obj: Any,
    include: Optional[IncEx] = None,
    exclude: Optional[IncEx] = None,
    by_alias: bool = True,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
    custom_encoder: Optional[Dict[Any, Callable[[Any], Any]]] = None,
    sqlalchemy_safe: bool = True,
) -> Any:
    custom_encoder = custom_encoder or {}
    if custom_encoder:
        if type(obj) in custom_encoder:
            return custom_encoder[type(obj)](obj)
        else:
            for encoder_type, encoder_instance in custom_encoder.items():
                if isinstance(obj, encoder_type):
                    return encoder_instance(obj)
    if include is not None and not isinstance(include, (set, dict)):
        include = set(include)
    if exclude is not None and not isinstance(exclude, (set, dict)):
        exclude = set(exclude)
    if isinstance(obj, BaseModel):
        # TODO: remove when deprecating Pydantic v1
        encoders: Dict[Any, Any] = {}
        if not PYDANTIC_V2:
            encoders = getattr(obj.__config__, "json_encoders", {})  # type: ignore[attr-defined]
            if custom_encoder:
                encoders.update(custom_encoder)
        obj_dict = _model_dump(
            obj,
            mode="json",
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )
        if "__root__" in obj_dict:
            obj_dict = obj_dict["__root__"]
        return jsonable_encoder(
            obj_dict,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
            # TODO: remove when deprecating Pydantic v1
            custom_encoder=encoders,
            sqlalchemy_safe=sqlalchemy_safe,
        )
    if dataclasses.is_dataclass(obj):
        obj_dict = dataclasses.asdict(obj)  # type: ignore
        return jsonable_encoder(
            obj_dict,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            custom_encoder=custom_encoder,
            sqlalchemy_safe=sqlalchemy_safe,
        )
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, PurePath):
        return str(obj)
    if isinstance(obj, (str, int, float, type(None))):
        return obj
    if isinstance(obj, dict):
        encoded_dict = {}
        allowed_keys = set(obj.keys())
        if include is not None:
            allowed_keys &= set(include)
        if exclude is not None:
            allowed_keys -= set(exclude)
        for key, value in obj.items():
            if (
                (
                    not sqlalchemy_safe
                    or (not isinstance(key, str))
                    or (not key.startswith("_sa"))
                )
                and (value is not None or not exclude_none)
                and key in allowed_keys
            ):
                encoded_key = jsonable_encoder(
                    key,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_none=exclude_none,
                    custom_encoder=custom_encoder,
                    sqlalchemy_safe=sqlalchemy_safe,
                )
                encoded_value = jsonable_encoder(
                    value,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_none=exclude_none,
                    custom_encoder=custom_encoder,
                    sqlalchemy_safe=sqlalchemy_safe,
                )
                encoded_dict[encoded_key] = encoded_value
        return encoded_dict
    if isinstance(obj, (list, set, frozenset, GeneratorType, tuple, deque)):
        encoded_list = []
        for item in obj:
            encoded_list.append(
                jsonable_encoder(
                    item,
                    include=include,
                    exclude=exclude,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    exclude_none=exclude_none,
                    custom_encoder=custom_encoder,
                    sqlalchemy_safe=sqlalchemy_safe,
                )
            )
        return encoded_list

    if type(obj) in ENCODERS_BY_TYPE:
        return ENCODERS_BY_TYPE[type(obj)](obj)
    for encoder, classes_tuple in encoders_by_class_tuples.items():
        if isinstance(obj, classes_tuple):
            return encoder(obj)

    try:
        data = dict(obj)
    except Exception as e:
        errors: List[Exception] = []
        errors.append(e)
        try:
            data = vars(obj)
        except Exception as e:
            errors.append(e)
            # keep the ValueError used by original FastAPI module
            raise ValueError(errors) from e
    return jsonable_encoder(
        data,
        include=include,
        exclude=exclude,
        by_alias=by_alias,
        exclude_unset=exclude_unset,
        exclude_defaults=exclude_defaults,
        exclude_none=exclude_none,
        custom_encoder=custom_encoder,
        sqlalchemy_safe=sqlalchemy_safe,
    )



# coding=utf-8
import json

from aliyunsdkcore.client import AcsClient
from aliyunsdksts.request.v20150401.AssumeRoleRequest import AssumeRoleRequest
from pydantic import BaseModel

from ..typings import PydanticModelConfig
from .base import Aliyun


class AliyunStsToken(BaseModel):
    model_config = PydanticModelConfig.default()
    access_key_secret: str
    security_token: str
    access_key_id: str


class AliyunSts(Aliyun):
    def __init__(self, region, **kwargs):
        super().__init__(**kwargs)
        self.__region = region

    def get_sts(self, session_name: str, role_arn: str) -> AliyunStsToken:
        # 构建一个阿里云客户端，用于发起请求。
        # 设置调用者（RAM用户或RAM角色）的AccessKey ID和AccessKey Secret。
        client = AcsClient(self.access_key, self.secret, self.__region)

        request = AssumeRoleRequest()
        request.set_accept_format("json")

        # set role
        request.set_RoleArn(role_arn)

        # give a name to this session, this is depends on you. no strict requirement
        request.set_RoleSessionName(session_name)

        response = client.do_action_with_exception(request)
        response_dict = json.loads(response.decode("utf-8"))["Credentials"]  # type: ignore
        return AliyunStsToken(
            access_key_secret=response_dict["AccessKeySecret"],
            security_token=response_dict["SecurityToken"],
            access_key_id=response_dict["AccessKeyId"],
        )



import logging
import time
from datetime import datetime, timedelta
from threading import Semaphore, Thread
from typing import Any, Callable, Dict, List, Optional, TypedDict

OwnerLock = TypedDict(
    "OwnerLock", {"acquiring_lock": List[int], "sema": Semaphore, "__created": datetime}
)

logger = logging.getLogger(__name__)


class ThreadController:
    def __init__(self, thread_size: int):
        self.sema = Semaphore(thread_size)
        self.thread_size = thread_size

        # a lock/queue for task that are acquiring semaphore
        self.acquiring_lock: List[int] = []
        self.owners: Dict[str | int, OwnerLock] = {}

    def create_thread(
        self,
        target: Callable,
        *,
        args: Optional[tuple] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        thread_owner: Optional[str | int] = None,
        disable_sema: Optional[bool] = False,
        join: Optional[bool] = False,
    ):
        # prevent memory leak
        # clear outdated owners
        self.clear_outdated_owners()
        self.acquiring_lock.append(1)  # add to queue
        self.init_owner(thread_owner=thread_owner)  # init owner
        self._acquire_owner_lock(thread_owner)  # get owner lock
        if not disable_sema:
            # todo
            #   Check is this double acquire compare to below
            self.acquire(thread_owner=thread_owner)
            self._release_owner_lock(thread_owner)
            self.acquiring_lock.pop()  # release the acquire lock
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = {}
        thread = Thread(target=target, args=args, kwargs=kwargs)
        thread.start()
        if join:
            # todo
            #   Check is this double acquire compare to above
            self.acquire(thread_owner=thread_owner)
            if not disable_sema:
                self.release(thread_owner=thread_owner)
        return thread

    def acquire(self, *, thread_owner: Optional[str | int] = None):
        self.sema.acquire()
        self._acquire_owner_sema(thread_owner)

    def release(self, *, thread_owner: Optional[str | int] = None):
        self.sema.release()
        self._release_owner_sema(thread_owner)

    def wait_all_to_finish(self):
        """Check is all the thread completely by acquire"""
        slept_secs = 0
        # wait until no task are asking for semaphore
        max_wait = 12 * 60 * 60  # 12 hours
        while len(self.acquiring_lock) > 0:
            time.sleep(1)
            slept_secs += 1
            if slept_secs % 300 == 0:
                if slept_secs >= max_wait:
                    # use ValueError for generalization
                    raise ValueError("Max iteration reached")

        for i in range(self.thread_size):
            self.sema.acquire()
        self.sema.release(n=self.thread_size)

    def wait_owner_to_finish(self, thread_owner: str | int):
        """Check is all the thread of same owner completely by acquire"""
        assert thread_owner is not None, "[Helpers] You must provide a thread owner"
        slept_secs = 0
        # wait until no task are asking for semaphore
        max_wait = 1 * 60 * 60  # 1 hour
        while (
            thread_owner in self.owners
            and len(self.owners[thread_owner]["acquiring_lock"]) > 0
        ):
            time.sleep(1)
            slept_secs += 1
            if slept_secs % 300 == 0:
                if slept_secs >= max_wait:
                    raise ValueError("Max iteration reached")

        for i in range(self.thread_size):
            if thread_owner not in self.owners:
                break
            self.owners[thread_owner]["sema"].acquire()
        if thread_owner in self.owners:
            self.owners[thread_owner]["sema"].release(n=self.thread_size)
            # delete owner
            del self.owners[thread_owner]

    def clear_outdated_owners(self):
        """remove outdated owners from hash map"""
        threshold = datetime.now() - timedelta(hours=2)
        keys_to_del = []
        for owner, owner_dict in self.owners.items():
            if owner_dict["__created"] < threshold:
                keys_to_del.append(owner)
        for key in keys_to_del:
            if key in self.owners:
                del self.owners[key]

    def init_owner(self, thread_owner: Optional[str | int] = None):
        if thread_owner is None:
            return
        self.owners.setdefault(
            thread_owner,
            {
                "acquiring_lock": [],
                "__created": datetime.now(),
                "sema": Semaphore(self.thread_size),
            },
        )

    def _acquire_owner_lock(self, thread_owner: Optional[str | int] = None):
        if thread_owner is None or thread_owner not in self.owners:
            return
        self.owners[thread_owner]["acquiring_lock"].append(1)

    def _release_owner_lock(self, thread_owner: Optional[str | int] = None):
        if thread_owner is None or thread_owner not in self.owners:
            return
        self.owners[thread_owner]["acquiring_lock"].pop()

    def _acquire_owner_sema(self, thread_owner: Optional[str | int] = None):
        if thread_owner is None or thread_owner not in self.owners:
            return
        self.owners[thread_owner]["sema"].acquire()

    def _release_owner_sema(self, thread_owner: Optional[str | int] = None):
        if thread_owner is None or thread_owner not in self.owners:
            return
        self.owners[thread_owner]["sema"].release()


def threaded(*, sema: Optional[Semaphore] = None, join: Optional[bool] = False):
    def threaded_func(func):
        """
        Decorator that multithreads the target function
        with the given parameters. Returns the thread
        created for the function
        """

        def wrapper(*args, **kwargs):
            # acquire sema if sema is provided
            if sema is not None:
                sema.acquire()
            try:
                thread = Thread(target=func, args=args, kwargs=kwargs)
                thread.start()
                if join:
                    thread.join()  # join if sema is provided, and then release
                    if sema is not None:
                        sema.release()
            except Exception as e:
                logger.critical(f"[Helper] Failed to run thread due to exception {e}")
                raise e  # bubble up the exception
            return thread

        return wrapper

    return threaded_func



from enum import Enum
from typing import Any, Literal, Optional

from pydantic import BaseModel, Field

from ..typings import PydanticModelConfig


class AgenticActionWarningLevel(str, Enum):
    SAFE = "SAFE"
    MUTATION = "MUTATION"
    DESTRUCTION = "DESTRUCTION"


class AgenticActionPayload(BaseModel):
    model_config = PydanticModelConfig.default()
    message: str = Field(
        description="A simple, human-friendly message explaining why you want to do this action (used for confirmation). Don’t include IDs, keys, or technical terms due to most user may be complete out-of-context and non-developer."
    )
    warning_level: AgenticActionWarningLevel = Field(
        description="The level of warning associated with the action; When no data mutation/or descruction is involved, use SAFE; otherwise, use MUTATION or DESTRUCTION accordingly"
    )
    object: Literal["agentic_action_payload"] = Field(
        default="agentic_action_payload",
        description="A string representing the object type.",
    )

    @classmethod
    def extract_agentic_action_payload(
        cls, args: Any
    ) -> Optional["AgenticActionPayload"]:
        def _normalize_candidate(candidate: Any) -> dict[str, Any] | None:
            if not isinstance(candidate, dict):
                return None
            if (
                "object" not in candidate
                or candidate["object"] != "agentic_action_payload"
            ):  # a literal string name in the dict, to identify the object type
                return None
            if "message" not in candidate:
                raise ValueError("agentic_action_payload missing 'message' field")
            if "warning_level" in candidate:
                warning_level = candidate["warning_level"]
            elif "warningLevel" in candidate:
                warning_level = candidate["warningLevel"]
            else:
                raise ValueError(
                    "agentic_action_payload missing 'warning_level'/'warningLevel' field"
                )
            return {"message": candidate["message"], "warning_level": warning_level}

        def _find_candidate(value: Any) -> dict[str, Any] | None:
            normalized = _normalize_candidate(value)
            if normalized is not None:
                return normalized
            if isinstance(value, dict):
                for v in value.values():
                    found = _find_candidate(v)
                    if found is not None:
                        return found
            elif isinstance(value, list):
                for v in value:
                    found = _find_candidate(v)
                    if found is not None:
                        return found
            return None

        candidate = _find_candidate(args)
        if candidate is None:
            return None
        return AgenticActionPayload(**candidate)


class AgentDeferredReason(BaseModel):
    model_config = PydanticModelConfig.default()
    reasons: list[str] = Field(
        default_factory=list, description="The reasons of the deferred tool request."
    )


class AgentDeferredToolFormDescription(BaseModel):
    model_config = PydanticModelConfig.default()
    markdown: Optional[str] = Field(default=None)


class AgentDeferredToolFormLabel(BaseModel):
    model_config = PydanticModelConfig.default()
    label: Optional[str] = Field(default=None)
    description: Optional[AgentDeferredToolFormDescription] = Field(default=None)


class AgentDeferredToolFormItemTextInput(BaseModel):
    model_config = PydanticModelConfig.default()
    placeholder: Optional[str] = Field(default=None)
    value: Optional[str] = Field(default=None)
    max_len: Optional[int] = Field(default=None)
    min_len: Optional[int] = Field(default=None)


class AgentDeferredToolFormItemTextListInput(BaseModel):
    model_config = PydanticModelConfig.default()
    placeholder: Optional[str] = Field(default=None)
    value: Optional[list[str]] = Field(default=None)
    str_max_len: Optional[int] = Field(default=None)
    str_min_len: Optional[int] = Field(default=None)
    list_max_len: Optional[int] = Field(default=None)
    list_min_len: Optional[int] = Field(default=None)


class AgentDeferredToolFormItemNumberInput(BaseModel):
    model_config = PydanticModelConfig.default()
    placeholder: Optional[str] = Field(default=None)
    value: Optional[float] = Field(default=None)
    gt: Optional[float] = Field(default=None)
    lt: Optional[float] = Field(default=None)
    gte: Optional[float] = Field(default=None)
    lte: Optional[float] = Field(default=None)


class AgentDeferredToolFormItemSelectOption(BaseModel):
    model_config = PydanticModelConfig.default()
    label_markdown: Optional[str] = Field(default=None)
    value: str | float


class AgentDeferredToolFormItemMultiSelect(BaseModel):
    model_config = PydanticModelConfig.default()
    options: list[AgentDeferredToolFormItemSelectOption] = Field(default_factory=list)
    values: Optional[list[str | float]] = Field(
        default=None, description="filled by user, used for results form"
    )


class AgentDeferredToolFormItemSingleSelect(BaseModel):
    model_config = PydanticModelConfig.default()
    options: list[AgentDeferredToolFormItemSelectOption] = Field(default_factory=list)
    value: Optional[str | float] = Field(
        default=None, description="filled by user, used for results form"
    )


class AgentDeferredToolFormItemCheckbox(BaseModel):
    model_config = PydanticModelConfig.default()
    checked_markdown: Optional[str] = Field(default=None)
    unchecked_markdown: Optional[str] = Field(default=None)
    value: Optional[bool] = Field(default=None)


class AgentDeferredToolFormItemType(str, Enum):
    TEXTAREA = "TEXTAREA"
    TEXT_INPUT = "TEXT_INPUT"
    TEXT_LIST_INPUT = "TEXT_LIST_INPUT"
    INT_INPUT = "INT_INPUT"
    FLOAT_INPUT = "FLOAT_INPUT"
    SINGLE_SELECT = "SINGLE_SELECT"
    MULTI_SELECT = "MULTI_SELECT"
    CHECKBOX = "CHECKBOX"


class AgentDeferredToolFormItem(AgentDeferredReason):
    model_config = PydanticModelConfig.default()
    item_type: AgentDeferredToolFormItemType
    label: AgentDeferredToolFormLabel
    text_input: Optional[AgentDeferredToolFormItemTextInput] = Field(
        default=None, description="For item of type TEXTAREA or TEXT_INPUT"
    )
    text_list_input: Optional[AgentDeferredToolFormItemTextListInput] = Field(
        default=None, description="For item of type TEXT_LIST_INPUT"
    )
    number_input: Optional[AgentDeferredToolFormItemNumberInput] = Field(
        default=None, description="For item of type INT_INPUT or FLOAT_INPUT"
    )
    single_select: Optional[AgentDeferredToolFormItemSingleSelect] = Field(
        default=None, description="For item of type SINGLE_SELECT"
    )
    multi_select: Optional[AgentDeferredToolFormItemMultiSelect] = Field(
        default=None, description="For item of type MULTI_SELECT"
    )
    checkbox: Optional[AgentDeferredToolFormItemCheckbox] = Field(
        default=None, description="For item of type CHECKBOX"
    )

    @property
    def value(self) -> Any:
        if self.item_type in {
            AgentDeferredToolFormItemType.TEXTAREA,
            AgentDeferredToolFormItemType.TEXT_INPUT,
        }:
            return self.text_input.value if self.text_input else None
        elif self.item_type == AgentDeferredToolFormItemType.TEXT_LIST_INPUT:
            return self.text_list_input.value if self.text_list_input else None
        elif self.item_type in {
            AgentDeferredToolFormItemType.INT_INPUT,
            AgentDeferredToolFormItemType.FLOAT_INPUT,
        }:
            return self.number_input.value if self.number_input else None
        elif self.item_type == AgentDeferredToolFormItemType.SINGLE_SELECT:
            return self.single_select.value if self.single_select else None
        elif self.item_type == AgentDeferredToolFormItemType.MULTI_SELECT:
            return self.multi_select.values if self.multi_select else None
        elif self.item_type == AgentDeferredToolFormItemType.CHECKBOX:
            return self.checkbox.value if self.checkbox else None
        return None


class AgentDeferredToolRequestsForm(AgentDeferredReason):
    model_config = PydanticModelConfig.default()
    title: Optional[str] = Field(default=None)
    description: Optional[AgentDeferredToolFormDescription] = Field(default=None)
    inputs: list[AgentDeferredToolFormItem]


class AgentDeferredToolResultsForm(AgentDeferredReason):
    model_config = PydanticModelConfig.default()
    inputs: list[AgentDeferredToolFormItem] = Field(default_factory=list)


class AgentDeferredToolPayloadBase(AgentDeferredReason):
    model_config = PydanticModelConfig.default()

    # metadata within metadata
    metadata: Optional[dict[str, Any]] = Field(
        default=None,
        description="Optional metadata associated with the deferred tool request.",
    )
    form: Optional[AgentDeferredToolRequestsForm | AgentDeferredToolResultsForm]

    def get_form_value_by_deferred_reason(self, reason: str) -> Any:
        if self.form is None:
            return None
        for item in self.form.inputs:
            if reason in item.reasons:
                return item.value
        return None


class AgentDeferredToolRequestsPayload(AgentDeferredToolPayloadBase):
    model_config = PydanticModelConfig.default()

    question: Optional[str] = Field(
        default=None,
        description="The question that require user to answer before the agent can continue the action. This is used for deferred tool request.",
    )
    default_answer: Optional[str] = Field(
        default=None,
        description="An optional default answer to suggest to the user.",
    )
    form: Optional[AgentDeferredToolRequestsForm] = Field(
        default=None,
        description="The form that the user need to fill out to answer the question.",
    )


class AgentDeferredToolResultsPayload(AgentDeferredToolPayloadBase):
    model_config = PydanticModelConfig.default()
    answer: Optional[str] = Field(
        default=None,
        description="The answer provided by the user to response the  deferred tool request's question.",
    )
    form: Optional[AgentDeferredToolResultsForm] = Field(
        default=None,
        description="The form filled out by the user to response the deferred tool request's form.",
    )



import logging

from redis import Redis

from .schemas import ProgressReport, ProgressTask

logger = logging.getLogger(__name__)


class ProgressHandler:
    def __init__(self, redis: Redis):
        self._redis = redis

    def new(self, *, key: str, model_key: str | None = None) -> ProgressTask:
        """
        Build a progress task object, you can provide 2 key
        :param key: to save this task progress, key can be used to retrieve progress percentage
        :param model_key:
        :return:
        """

        root = ProgressTask(redis=self._redis, key=key, model_key=model_key)
        if model_key is not None:
            prebuilt_tasks = ProgressTask.get_prebuilt_subtasks_model(
                root=root, redis=self._redis, model_key=model_key
            )
            root.subtasks = prebuilt_tasks
        logger.info(
            f"[Helper] Progress task inited, and initialized state: {str(root)}"
        )
        return root

    def get(self, *, key: str) -> ProgressReport | None:
        t = ProgressTask.get_from_redis(redis=self._redis, key=key)
        if t is None:
            return None
        return t.progress_report



import re
from typing import List

import tiktoken


def count_tokens(string: str, encoding_name: str = "cl100k_base") -> int:
    """Return numbers of token"""
    # https://stackoverflow.com/questions/75804599/openai-api-how-do-i-count-tokens-before-i-send-an-api-request
    # https://github.com/openai/tiktoken
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens


def split_string_into_slices(s: str, slice_size: int) -> List[str]:
    return [s[i : i + slice_size] for i in range(0, len(s), slice_size)]


def limit_tokens(content: str, max_token: int) -> str:
    slice_size = 10
    split_ = split_string_into_slices(content, slice_size)
    last_string = ""
    for sp in split_:
        new_token = count_tokens(last_string + sp)
        if new_token < max_token:
            last_string += sp
        else:
            break
    return last_string


def escape_json_string_values(json_string: str) -> str:
    """escape \\n and \\t in json string
    # credit: chatgpt
    :param json_string:
    :return:
    """

    def escape_special_chars(match):
        inner_str = match.group(1)
        # Escape backslashes first to avoid double escaping
        inner_str = inner_str.replace("\\", "\\\\")
        inner_str = inner_str.replace("\n", "\\n").replace("\t", "\\t")
        return f'"{inner_str}"'

        # Use regex to find all string values in JSON and apply escaping

    escaped_json_string = re.sub(
        r'"([^"\\]*(?:\\.[^"\\]*)*)"', escape_special_chars, json_string
    )

    return escaped_json_string



import hashlib
import os
import random
import shutil
import string
import uuid
import zipfile
from datetime import date, datetime
from pathlib import Path
from types import GeneratorType
from typing import Any, Iterator, List

import keble_exceptions
from pydantic import BaseModel


def id_generator() -> str:
    return str(uuid.uuid4())


def stable_id_generator(
    *,
    payload: dict | BaseModel,  # if the payload unchanged, the id will be the same
    allow_timestamp_keys: bool = False,
) -> str:
    """Generate a stable UUID5 based on the provided payload dictionary (excluding timestamp)."""
    forbidden_keys = [
        "created",
        "updated",
        "created_at",
        "updated_at",
        "createdAt",
        "updatedAt",
        "created_ts",
        "updated_ts",
        "createdTs",
        "updatedTs",
    ]
    if isinstance(payload, BaseModel):
        payload = payload.model_dump(
            exclude=set(forbidden_keys) if not allow_timestamp_keys else None
        )
    if not allow_timestamp_keys:
        for key in forbidden_keys:
            if key in payload:
                raise ValueError(
                    f"[Helpers] Payload dictionary must not contain timestamp key: {key}"
                )

    seed_list = [f"{key}:{val}" for key, val in sorted(payload.items())]
    seed = "|".join(seed_list)
    digest = hashlib.sha1(seed.encode("utf-8")).hexdigest()
    return str(uuid.uuid5(uuid.NAMESPACE_DNS, f"{digest}"))


def generate_random_string(
    length: int = 32, *, lower: bool = True, upper: bool = True, digit: bool = True
) -> str:
    candidates = []
    if lower:
        candidates += string.ascii_lowercase
    if upper:
        candidates += string.ascii_uppercase
    if digit:
        candidates += string.digits
    assert len(candidates) > 0, "Invalid random string generator, missing candidates."
    return "".join(random.choice(candidates) for i in range(length))


def is_pydantic_field_empty(obj: BaseModel, field: str) -> bool:
    return not hasattr(obj, field) or getattr(obj, field) is None


def date_to_datetime(d: date) -> datetime:
    if isinstance(d, date):
        return datetime.combine(d, datetime.min.time())
    return d


def datetime_to_date(d: datetime) -> date:
    if isinstance(d, datetime):
        return d.date()
    return d


# _hashu = lambda word: ctypes.c_uint64(hash(word)).value


def hash_string(arg: str) -> str:
    hash_object = hashlib.md5(arg.encode())
    return hash_object.hexdigest()

    # return str(hash(args))
    # return _hashu("".join(args)).to_bytes(8, "big").hex()


def slice_to_list(items: List[Any], slice_size: int) -> List[List[Any]]:
    """Convert 1D list to 2D list"""
    slices: List[List[Any]] = []
    current_slice: List[Any] = []
    for _d in items:
        current_slice.append(_d)
        if len(current_slice) >= slice_size:
            slices.append(current_slice)
            current_slice = []
    if len(current_slice) > 0:
        slices.append(current_slice)
    return slices


def bad_utf8_str_encoding(str_: str) -> str:
    return repr(str_)[1:-1]


def get_first_match(items: list, key_fn, value):
    """Return first match item in a list"""
    filtered = list(filter(lambda x: key_fn(x) == value, items))
    if len(filtered) >= 1:
        return filtered[0]
    # return None for not found
    return None


def ensure_has_folder(path: str) -> str:
    exist = os.path.exists(path)
    if not exist:
        # Create a new directory because it does not exist
        os.makedirs(path)
    return path


def yield_files(folder: str) -> Iterator[str | Path]:
    for filename in os.listdir(folder):
        f = os.path.join(folder, filename)
        if os.path.isfile(f):
            yield f
        elif os.path.isdir(f):
            yield from yield_files(f)


def get_files(folder: str) -> List[str | Path]:
    all_files = []
    for filename in os.listdir(folder):
        f = os.path.join(folder, filename)
        if os.path.isfile(f):
            all_files.append(f)
        elif os.path.isdir(f):
            all_files += get_files(f)
    return all_files


def zip_dir(folder: Path | str, zip_filepath: Path | str):
    """Zip the provided directory without navigating to that directory using `pathlib` module"""

    # Convert to Path object
    dir = Path(folder)

    with zipfile.ZipFile(zip_filepath, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for entry in dir.rglob("*"):
            zip_file.write(entry, entry.relative_to(dir))


def remove_dir(dir: Path | str):
    shutil.rmtree(dir, ignore_errors=True)


def wait_generator_stop(generator: GeneratorType, *, max_generate: int = 1000):
    generated = []
    attempts = 0
    while True:
        try:
            next_item = next(generator)
            generated.append(next_item)
        except StopIteration:
            break
        attempts += 1
        if attempts > max_generate:
            raise keble_exceptions.MaxIterationReached(
                admin_note=f"Generator exceeded max generate allowance {max_generate}",
                alert_admin=True,
            )
            # raise ValueError(f"Generator exceeded max generate allowance {max_generate}")


def is_mime_prefix_in(mime, mime_start: List[str]):
    assert mime is not None and "/" in mime, f"Invalid mime found: {mime}"
    start = mime.split("/")[0]
    return start in mime_start


def is_mime_image(mime: str):
    return is_mime_prefix_in(mime, ["image"])


video_mimes = [
    "video/x-msvideo",  # .avi
    "video/mp4",  # .mp4, .m4v
    "video/mpeg",  # .mpeg, .mpg
    "video/quicktime",  # .mov, .qt
    "video/vnd.vivo",  # .viv, .vivo
    "video/x-ms-wmv",  # .wmv
    "video/x-ms-asf",  # .asf, .asx
    "video/x-ms-wm",  # .wm
    "video/x-ms-wmx",  # .wmx
    "video/x-ms-wvx",  # .wvx
    "video/3gpp",  # .3gp, .3gpp
    "video/3gpp2",  # .3g2
    "video/h263",  # .h263
    "video/webm",  # .webm
    "video/x-matroska",  # .mkv
    "video/x-flv",  # .flv
    "video/ogg",  # .ogg, .ogv
    "video/h264",  # .h264
    "video/h265",  # .h265, .hevc
    "video/x-m4v",  # .m4v
    "application/x-troff-msvideo",  # .avi
    "video/avi",  # .avi
    "video/msvideo",  # .avi
    "video/x-dv",  # .dv
    "video/x-ivf",  # .ivf
    "video/x-la-asf",  # .lsf, .lsx
    "video/x-mng",  # .mng
    "video/x-ms-asf-plugin",  # .asf
    "video/x-ms-vob",  # .vob
    "video/x-mpg",  # .mpg
    "video/x-mpeg",  # .mpeg
    "video/x-mpeg2a",  # .mp2
    "video/x-pv-pvx",  # .pvx
    "video/x-qtc",  # .qtc
    "video/x-sgi-movie",  # .movie
]

audio_mimes = [
    "audio/3gpp",  # .3ga
    "audio/8svx",  # .8svx
    "audio/aac",  # .aac
    "audio/ac3",  # .ac3
    "audio/aiff",  # .aif, .aiff
    "audio/alac",  # .alac
    "audio/amr",  # .amr
    "audio/ape",  # .ape
    "audio/basic",  # .au
    "audio/x-dss",  # .dss
    "audio/flac",  # .flac
    "audio/x-flv",  # .flv
    "audio/mp4",  # .m4a, .m4b, .m4p, .m4r
    "audio/mpeg",  # .mp3, .mpga
    "audio/ogg",  # .ogg, .oga, .mogg
    "audio/opus",  # .opus
    "audio/qcelp",  # .qcp
    "audio/x-tta",  # .tta
    "audio/voc",  # .voc
    "audio/wav",  # .wav
    "audio/x-ms-wma",  # .wma
    "audio/wavpack",  # .wv
]

ms_excel_mimes = [
    "application/vnd.ms-excel",  # .xls
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",  # .xlsx
    "application/vnd.openxmlformats-officedocument.spreadsheetml.template",  # .xltx
    "application/vnd.ms-excel.sheet.macroEnabled.12",  # .xlsm
    "application/vnd.ms-excel.template.macroEnabled.12",  # .xltm
    "application/vnd.ms-excel.addin.macroEnabled.12",  # .xlam
    "application/vnd.ms-excel.sheet.binary.macroEnabled.12",  # .xlsb
]


def is_mime_video(mime: str):
    return is_mime_prefix_in(mime, ["video"]) or mime in video_mimes


def is_mime_audio(mime: str):
    return is_mime_prefix_in(mime, ["audio"]) or mime in audio_mimes


def is_mime_media(mime: str):
    return is_mime_prefix_in(mime, ["image", "video", "audio"])


def is_mime_ms_excel(mime: str):
    return mime in ms_excel_mimes


def is_mime_csv(mime: str):
    return mime == "text/csv"


def inline_string(string: str, max_len: int = 30):
    if string is None:
        return "<None Value>"
    s = string.replace("\n", "").replace("\t", "")
    if len(s) > max_len:
        return " < " + s[0:max_len] + "... > "
    else:
        return " < " + s + " > "



from pydantic import ConfigDict
from pydantic.alias_generators import to_camel


class PydanticModelConfig:
    @classmethod
    def default_dict(cls, **kwargs) -> dict:
        return {"populate_by_name": True, "alias_generator": to_camel, **kwargs}

    @classmethod
    def default(cls, **kwargs) -> ConfigDict:
        return ConfigDict(**cls.default_dict(**kwargs))



import json
import logging
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, ConfigDict
from redis import Redis

from ..typings.config import PydanticModelConfig

logger = logging.getLogger(__name__)


class ProgressTaskStage(str, Enum):
    PENDING = "PENDING"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class ProgressReport(BaseModel):
    model_config = PydanticModelConfig.default()
    progress_key: Optional[str] = None
    progress: float
    is_root_success: bool
    success: int
    failure: int
    pending: int
    assigned: int
    total: int
    message: Optional[str] = None
    errors: List[str] = []


class ProgressTask(BaseModel):
    model_config = ConfigDict(
        **PydanticModelConfig.default_dict(), arbitrary_types_allowed=True
    )
    stage: ProgressTaskStage = ProgressTaskStage.PENDING
    subtasks: List["ProgressTask"] = []
    error: Optional[str] = None
    redis: Optional[Redis] = None
    key: Optional[str] = None
    model_key: Optional[str] = None
    root: Optional["ProgressTask"] = None
    message: Optional[str] = None
    # if the task is assigned to a task
    # it will be marked as True
    # otherwise, it will be marked as False (by default)
    assigned: bool = False

    # @property
    # def root_key(self) -> str:
    #     if self.root is not None:
    #         return self.root.root_key
    #     return self.key

    def __str__(self):
        report = self.progress_report
        return f"[Helper] Progress {f'{round(report.progress * 100, 2)}%' if report.progress is not None else '-%'}(total={report.total}, assigned={report.assigned}, success={report.success}, failure={report.failure}, pending={report.pending})"

    @property
    def is_root_success(self) -> bool:
        if self.root is not None:
            return self.root.is_root_success
        # self is root
        return self.stage == ProgressTaskStage.SUCCESS

    @property
    def total_assigned(self) -> int:
        if self.root is not None:
            return self.root.total_assigned
        # is root, 1 is for root, root is assumed as assigned
        return 1 + len([s for s in self.subtasks if s.assigned is True])

    @classmethod
    def get_from_redis(cls, redis: Redis, *, key: str) -> Optional["ProgressTask"]:
        h = redis.get(key)
        if h is None:
            return None
        return ProgressTask(**json.loads(h))  # type: ignore

    @classmethod
    def get_prebuilt_subtasks_model(
        cls, root: "ProgressTask", redis: Redis, *, model_key: str
    ) -> List["ProgressTask"]:
        """number of subtasks will be stored in redis,
        so each time when a new object is initialized,
         a certain amount of subtasks will be prebuilt into the list"""
        h = redis.get(model_key)
        if h is None:
            return []
        return [
            ProgressTask(
                root=root,
            )
            for _ in range(max(int(h) - 1, 0))  # type: ignore
        ]

    def set_message(self, message: Optional[str]):
        if self.root is not None:
            return self.root.set_message(message)
        self.message = message
        self._refresh_redis(update_model=self.is_root_success)

    def _refresh_redis(self, update_model: bool):
        if self.root is not None:
            self.root._refresh_redis(update_model=update_model)
        else:
            # now self is root
            assert self.key is not None and self.redis is not None, (
                "[Helpers] You must provide redis and key for progress task to refresh redis cache"
            )
            self.redis.set(
                self.key,
                json.dumps(self.model_dump_circular_reference_safe()),
                ex=48 * 60 * 60,  # expire after 48 hours
            )
            if self.model_key is not None and update_model:
                # we will update how many subtask it requires to complete
                logger.debug(
                    f"[Helper] Progress model update as total assigned={self.total_assigned}, model_key={self.model_key}"
                )
                self.redis.set(
                    self.model_key,
                    self.total_assigned,
                    ex=30 * 24 * 60 * 60,  # expire after 30 days
                )

    def new_subtask(self) -> "ProgressTask":
        # Avoid setting root to self, set it only to the rootmost instance
        if self.root is not None:
            return self.root.new_subtask()
        else:
            # is root, handle subtask by itself
            # before creating any new subtask
            # we first search unassigned task object
            unassigned_list = [s for s in self.subtasks if not s.assigned]
            if len(unassigned_list) > 0:
                subtask = unassigned_list[0]
            else:
                subtask = ProgressTask(root=self)
                self.subtasks.append(subtask)  # append this new subtask to list

            subtask.assigned = True  # mark as assigned
            self._refresh_redis(update_model=self.is_root_success)
            return subtask

    def success(self):
        self.stage = ProgressTaskStage.SUCCESS
        self._refresh_redis(update_model=self.is_root_success)

    def failure(self, error: Optional[str] = None):
        self.stage = ProgressTaskStage.FAILURE
        self.error = error
        # all pending subtask need to mark as failure
        for s in self.subtasks:
            if s.stage == ProgressTaskStage.PENDING:
                s.failure()
        self._refresh_redis(update_model=self.is_root_success)

    @property
    def progress_report(self) -> ProgressReport:
        if self.root is not None:
            return self.root.progress_report
        total = 1 + len(self.subtasks)
        success_counts = len(
            [
                s
                for s in [self] + self.subtasks
                if s.stage in [ProgressTaskStage.SUCCESS]
            ]
        )
        failure_counts = len(
            [
                s
                for s in [self] + self.subtasks
                if s.stage in [ProgressTaskStage.FAILURE]
            ]
        )
        pending_counts = len(
            [
                s
                for s in [self] + self.subtasks
                if s.stage in [ProgressTaskStage.PENDING]
            ]
        )
        errors = [
            s.error
            for s in [self] + self.subtasks
            if s.stage in [ProgressTaskStage.FAILURE]
            if s.error is not None
        ]
        progress_floats = success_counts / total if total > 0 else 0
        return ProgressReport(
            progress_key=self.key,
            progress=progress_floats,
            is_root_success=self.is_root_success,
            success=success_counts,
            failure=failure_counts,
            pending=pending_counts,
            assigned=self.total_assigned,
            total=total,
            errors=errors,
            message=self.message,
        )

        # subtask_progress = [s.progress_report for s in self.subtasks]
        # total = 1 + len(self.subtasks)
        # progress_floats = [p.progress * (1 / total) for p in subtask_progress]
        #
        # success = sum([p.success for p in subtask_progress]) if len(subtask_progress) > 0 else 0
        # failure = sum([p.failure for p in subtask_progress]) if len(subtask_progress) > 0 else 0
        # pending = sum([p.pending for p in subtask_progress]) if len(subtask_progress) > 0 else 0
        # errors = reduce(lambda a, b: a + b, [p.errors for p in subtask_progress]) if len(subtask_progress) > 0 else []
        # if self.stage == ProgressTaskStage.SUCCESS:
        #     success += 1
        #     progress_floats.append(1 / total)
        # if self.stage == ProgressTaskStage.FAILURE:
        #     failure += 1
        #     if self.error is not None:
        #         errors.append(self.error)
        # if self.stage == ProgressTaskStage.PENDING:
        #     pending += 1

        # return ProgressReport(
        #     progress_key=self.key,
        #     progress=sum(progress_floats) if len(progress_floats) > 0 else 0,
        #     is_root_success=self.is_root_success,
        #     success=success,
        #     failure=failure,
        #     pending=pending,
        #     errors=errors
        # )

    def model_dump_circular_reference_safe(self) -> dict:
        """Custom model dump that excludes circular references."""
        return {
            "stage": self.stage,
            "error": self.error,
            "key": self.key,
            "message": self.message,
            "subtasks": [s.model_dump_circular_reference_safe() for s in self.subtasks],
        }
        # return self.model_dump(exclude={"root", "subtasks", "redis"}, mode="json")



from pydantic import BaseModel


class AliyunOssPutObjectResponse(BaseModel):
    # see https://help.aliyun.com/zh/oss/developer-reference/simple-upload-1?spm=a2c4g.11186623.0.i4
    status: int
    request_id: str
    etag: str
    headers: dict



from enum import Enum
from typing import Optional

from pydantic import BaseModel

from ..common import is_mime_csv, is_mime_image, is_mime_ms_excel, is_mime_video
from .config import PydanticModelConfig


class CloudStorageType(str, Enum):
    AWS_S3 = "AWS_S3"
    ALIYUN_OSS = "ALIYUN_OSS"


class CloudStorageObjectType(str, Enum):
    IMAGE = "IMAGE"
    VIDEO = "VIDEO"
    EXCEL = "EXCEL"
    CSV = "CSV"
    OTHER = "OTHER"

    @classmethod
    def determine_type(cls, *, mime: str) -> "CloudStorageObjectType":
        mime = mime.lower()
        if is_mime_video(mime):
            return CloudStorageObjectType.VIDEO
        if is_mime_image(mime):
            return CloudStorageObjectType.IMAGE
        if is_mime_ms_excel(mime):
            return CloudStorageObjectType.EXCEL
        if is_mime_csv(mime):
            return CloudStorageObjectType.CSV
        return CloudStorageObjectType.OTHER


class CloudStorageBase(BaseModel):
    """Provide a universal CloudStorage Class for different
    cloud service platform"""

    model_config = PydanticModelConfig.default()

    # key for AWS is straightforward, for aliyun, it is name/or oss path.
    # they all refer as key
    key: str

    # endpoint of the cloud storage
    # base_url + key should be the URL of the cloud storage
    base_url: str

    # type of cloud service
    type: CloudStorageType

    # type of file
    object_type: CloudStorageObjectType

    # original file name
    original_file_name: Optional[str]



from typing import Any
from urllib.parse import urlparse


def is_http_url(url: Any) -> bool:
    if url is None:
        return False
    val = str(url)
    parsed = urlparse(val)
    if parsed.scheme not in ("http", "https") or not parsed.netloc:
        return False
    return True



import math
from typing import Optional

import numpy as np


# You can use this helper function to check if a value is NaN or infinity
def is_invalid_float(value: Optional[float]) -> bool:
    if value is None:
        return False
    return math.isnan(value) or math.isinf(value)


def guard_invalid_float(value: float | None | np.floating) -> float | None:
    """
    Checks if the value is an invalid float (NaN, Infinity, -Infinity).
    If invalid, replaces it with None.

    Parameters:
    - value: The value to check, can be a float or a numpy float.

    Returns:
    - None if the value is invalid (NaN, inf, -inf), otherwise returns the value.
    """
    # Check for NaN or Infinity using numpy's functions
    if isinstance(value, np.floating):  # Handle np.float32, np.float64, etc.
        if np.isnan(value) or np.isinf(value):
            return None
        return float(value)
    elif isinstance(value, float):  # Handle Python native floats
        if np.isnan(value) or np.isinf(value):
            return None
    return value



# from .embedder import Embedder
from .agent import (
    AgentDeferredToolFormDescription,
    AgentDeferredToolFormItem,
    AgentDeferredToolFormItemCheckbox,
    AgentDeferredToolFormItemMultiSelect,
    AgentDeferredToolFormItemNumberInput,
    AgentDeferredToolFormItemSelectOption,
    AgentDeferredToolFormItemSingleSelect,
    AgentDeferredToolFormItemTextInput,
    AgentDeferredToolFormItemType,
    AgentDeferredToolFormLabel,
    AgentDeferredToolPayloadBase,
    AgentDeferredToolRequestsForm,
    AgentDeferredToolRequestsPayload,
    AgentDeferredToolResultsForm,
    AgentDeferredToolResultsPayload,
    AgenticActionPayload,
    AgenticActionWarningLevel,
)
from .helpers import (
    PromptUrlExtractionResult,
    convert_image_to_base64,
    extract_urls_from_prompt,
    is_image_url,
)
from .tokens import count_tokens, limit_tokens



import logging
import os.path
from pathlib import Path

import aiohttp
import keble_exceptions
from tenacity import retry, stop_after_attempt

logger = logging.getLogger(__name__)


@retry(stop=stop_after_attempt(3), reraise=True)
async def adownload_file(*, url: str, folder: Path, filename: str) -> Path:
    """Downloads an file from a URL and saves it to the specified folder."""
    folder.mkdir(parents=True, exist_ok=True)
    filepath = folder / filename
    if os.path.exists(filepath):
        # file already loaded and existed
        return filepath
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status == 200:
                with open(filepath, "wb") as f:
                    f.write(await response.read())
                return filepath
            else:
                raise keble_exceptions.RequestFailure(
                    admin_note={
                        "url": url,
                        "folder": str(folder),
                        "filename": filename,
                    },
                    alert_admin=True,
                    function_identifier="keble_helpers.download_file",
                )
                # raise Exception(
                #     f"[Helpers] Failed to download {url}: HTTP {response.status}"
                # )



import urllib.request
from typing import Optional

import oss2
from pydantic import AnyHttpUrl, BaseModel, Field

from ..typings import PydanticModelConfig
from .base import Aliyun
from .schemas import AliyunOssPutObjectResponse


# meta data for aliyun oss object, return from bucket.get_object_meta
class AliyunOssMeta(BaseModel):
    model_config = PydanticModelConfig.default()

    etag: Optional[str] = Field(None, description="Aliyun OSS etag")
    content_length: Optional[int] = Field(
        None, description="File size in bytes", serialization_alias="contentLength"
    )
    last_modified: Optional[int] = Field(
        None, description="File size in bytes", serialization_alias="lastModified"
    )
    content_type: Optional[str] = Field(
        None,
        description="MIME type of the file, e.g. img/jpeg",
        serialization_alias="contentType",
    )


class AliyunOss(Aliyun):
    # Aliyun OSS upload requires STS+RAM user authentication
    # see: https://help.aliyun.com/zh/oss/developer-reference/use-the-accesskey-pair-of-a-ram-user-to-initiate-a-request

    def __init__(self, oss_endpoint: AnyHttpUrl, bucket: str, **kwargs):
        super().__init__(**kwargs)
        self.__oss_endpoint: str = str(oss_endpoint).strip("/ ")
        self.__bucket_name: str = bucket.strip("/ ")

    def get_bucket(self) -> oss2.Bucket:
        auth = oss2.Auth(self.access_key, self.secret)
        return oss2.Bucket(auth, self.__oss_endpoint, self.__bucket_name)

    def get_bucket_with_sts(self, sts_token: str):
        auth = oss2.StsAuth(self.access_key, self.secret, sts_token)
        return oss2.Bucket(auth, self.__oss_endpoint, self.__bucket_name)

    def get_object_meta(self, key: str) -> AliyunOssMeta:
        """Get an oss file metadata
        example key as oss path: some_folder/file.txt"""
        bucket = self.get_bucket()
        meta = bucket.head_object(key)
        return AliyunOssMeta(
            etag=meta.etag,
            content_length=meta.content_length,
            last_modified=meta.last_modified,
            content_type=meta.content_type,
        )

    def save_object_to_local(self, key: str, local_path: str, *args, **kwargs):
        """Convert a file from oss path to local file
        key(as oss path) can be something like some_folder/file.txt"""
        bucket = self.get_bucket()
        bucket.get_object_to_file(key, local_path, *args, **kwargs)

    def save_local_to_cloud(self, key: str, local_path: str, *args, **kwargs):
        """Convert a file from oss path to local file
        key(as oss path) can be something like some_folder/file.txt"""
        bucket = self.get_bucket()
        res = bucket.put_object_from_file(key, local_path)
        return AliyunOssPutObjectResponse(
            status=res.status,
            request_id=res.request_id,
            etag=res.etag,  # type: ignore
            headers=res.headers,
        )

    def save_snapshot_to_local(self, key: str, local_path: str, seconds: int):
        """Get a snapshot file from an oss object.
        Use case: video snapshot"""
        bucket = self.get_bucket()

        signed_url = bucket.sign_url(
            "GET",
            key,
            10 * 60,
            params={
                "x-oss-process": f"video/snapshot,t_{seconds}000,f_jpg,w_0,h_0,m_fast"
            },
        )
        urllib.request.urlretrieve(signed_url, local_path)

    def cold_archive_object(self, key: str):
        """Convert a file into cold archive status"""
        bucket = self.get_bucket()

        # change object by changing its header class
        headers = {"x-oss-storage-class": oss2.BUCKET_STORAGE_CLASS_COLD_ARCHIVE}

        # override it by coping the object, from current path to the same current path
        bucket.copy_object(bucket.bucket_name, key, key, headers)

    def get_sts_signed_url(
        self,
        sts_token: str,
        key: str,
        *,
        expire_seconds: int = 60,
        content_type: Optional[str] = None,
        oss_storage_class: Optional[str] = None,
    ) -> str:
        bucket = self.get_bucket_with_sts(sts_token)

        # 指定Header。
        headers = dict()
        # You can set fixed Content-Type by:
        if content_type is not None:
            headers["Content-Type"] = content_type  # sample value: 'text/txt'

        # or configure storage class
        if oss_storage_class is not None:
            headers["x-oss-storage-class"] = (
                oss_storage_class  # sample value: "Standard"
            )

        # URL will expire after expire_seconds/60 seconds
        # due to OSS url error, you need to set slash_safe as TRUE
        url = bucket.sign_url(
            "GET", key, expire_seconds, slash_safe=True, headers=headers
        )
        return url



from . import np_utils
from .ai import *
from .aliyun import *
from .common import (
    audio_mimes,
    bad_utf8_str_encoding,
    date_to_datetime,
    datetime_to_date,
    ensure_has_folder,
    generate_random_string,
    get_files,
    get_first_match,
    hash_string,
    id_generator,
    inline_string,
    is_mime_audio,
    is_mime_csv,
    is_mime_image,
    is_mime_media,
    is_mime_ms_excel,
    is_mime_prefix_in,
    is_mime_video,
    is_pydantic_field_empty,
    ms_excel_mimes,
    remove_dir,
    slice_to_list,
    stable_id_generator,
    video_mimes,
    wait_generator_stop,
    yield_files,
    zip_dir,
)
from .datetime import days_in_month
from .enum import Environment
from .fastapi import jsonable_encoder
from .file import adownload_file
from .multithread import ThreadController, threaded
from .progress import *
from .typings import *
from .typings.standard_number import StdMoneyAmount, StdPercentage
from .typings.typing import (
    AmazonMarketplace,
    AmazonMarketplaceEU,
    Currency,
    ExchangeRateInUsd,
    Language,
    Marketplace,
    Money,
    MoneyRange,
    MongoObjectBase,
    NaturalLanguage,
    OneOrMoreVectors,
    OwnerType,
    PercentageRange,
    PromptRole,
    QdrantSchemaBase,
    SchemaBase,
    SharingScope,
    Status,
    StrapiSchemaBase,
    Timestamp,
    Vector,
    is_vectors_list,
)



from datetime import datetime, timezone
from enum import Enum, EnumMeta
import re
from typing import Any, List, Optional, Union
from urllib.parse import parse_qsl, urlsplit

import keble_exceptions
from pydantic import (
    AliasChoices,
    AnyHttpUrl,
    BaseModel,
    Field,
    GetCoreSchemaHandler,
    field_validator,
    model_validator,
)
from pydantic_core import CoreSchema, core_schema
from typing_extensions import Annotated

from keble_helpers.typings.id import ObjectId

from . import PydanticModelConfig
from .standard_number import StdMoneyAmount, StdPercentage

DEFAULT_MODEL_CONFIG = PydanticModelConfig.default()

__all__ = [
    "AmazonMarketplace",
    "AmazonMarketplaceEU",
    "Currency",
    "ExchangeRateInUsd",
    "Language",
    "Marketplace",
    "Money",
    "MoneyRange",
    "MongoObjectBase",
    "NaturalLanguage",
    "OneOrMoreVectors",
    "OwnerType",
    "PercentageRange",
    "PromptRole",
    "QdrantSchemaBase",
    "SchemaBase",
    "SharingScope",
    "Status",
    "StrapiSchemaBase",
    "Timestamp",
    "Vector",
    "is_vectors_list",
]

Vector = List[float]

OneOrMoreVectors = Vector | List[Vector]


def is_vectors_list(vectors: OneOrMoreVectors) -> bool:
    if vectors is None or len(vectors) == 0:
        raise keble_exceptions.ServerSideInvalidParams(
            invalid_params="vectors",
            admin_note=vectors,
            alert_admin=True,
            expected="not(vectors is None or len(vectors) == 0)",
            but_got=f"vectors={vectors}",
        )
        # raise ValueError("[Typing] Empty vectors")

    if isinstance(vectors, list) and len(vectors) > 0 and isinstance(vectors[0], list):
        return True
    return False


class _PromptRoleAliasEnumMeta(EnumMeta):
    # name alias in Enum class
    # credit: https://stackoverflow.com/questions/24105268/is-it-possible-to-override-new-in-an-enum-to-parse-strings-to-an-instance
    def __call__(cls, value, *args, **kw):
        if isinstance(value, str):
            # map strings to enum values
            if value == "ai":
                value = "assistant"
            if value == "system":
                value = "system"
            if value == "human":
                value = "user"
        return super().__call__(value, *args, **kw)


class PromptRole(str, Enum, metaclass=_PromptRoleAliasEnumMeta):
    """Prompt role enum in openai chatgpt fashion"""

    system = "system"
    user = "user"
    assistant = "assistant"


class Language(str, Enum):
    ENGLISH = "ENGLISH"
    SIMPLIFIED_CHINESE = "SIMPLIFIED_CHINESE"

    # extends by AI base on marketplace of amazon
    GERMAN = "GERMAN"
    FRENCH = "FRENCH"
    ITALIAN = "ITALIAN"
    SPANISH = "SPANISH"
    DUTCH = "DUTCH"
    PORTUGUESE = "PORTUGUESE"
    JAPANESE = "JAPANESE"
    TURKISH = "TURKISH"
    POLISH = "POLISH"
    SWEDISH = "SWEDISH"
    ARABIC = "ARABIC"
    HINDI = "HINDI"
    THAI = "THAI"
    KOREAN = "KOREAN"
    VIETNAMESE = "VIETNAMESE"
    INDONESIAN = "INDONESIAN"
    HEBREW = "HEBREW"
    NORWEGIAN = "NORWEGIAN"
    DANISH = "DANISH"
    FINNISH = "FINNISH"


class NaturalLanguage(BaseModel):
    """represent natural language"""

    model_config = DEFAULT_MODEL_CONFIG

    simplified_chinese: str
    english: str

    def get_text(self, language: Language) -> str:
        if language == Language.SIMPLIFIED_CHINESE:
            return self.simplified_chinese
        # by default, return english
        return self.english


class UnspecifiedMarketplace(str, Enum):
    GLOBAL = "GLOBAL"
    UNKNOWN = "UNKNOWN"


class Marketplace(str, Enum):
    CN = "CN"
    US = "US"
    CA = "CA"
    JP = "JP"
    FR = "FR"
    IT = "IT"
    DE = "DE"
    ES = "ES"
    UK = "UK"
    BR = "BR"
    BE = "BE"
    NL = "NL"
    SE = "SE"
    PL = "PL"
    EG = "EG"
    TR = "TR"
    SA = "SA"
    AE = "AE"
    IN = "IN"
    SG = "SG"
    AU = "AU"
    MX = "MX"

    @property
    def currency(self) -> "Currency":
        if self.value == Marketplace.US:
            return Currency.USD
        if self.value == Marketplace.CA:
            return Currency.CAD
        if self.value == Marketplace.JP:
            return Currency.JPY
        if self.value == Marketplace.UK:
            return Currency.GBP
        if self.value == Marketplace.SE:
            return Currency.SEK
        if self.value == Marketplace.PL:
            return Currency.PLN
        if self.value in [
            Marketplace.DE,
            Marketplace.ES,
            Marketplace.FR,
            Marketplace.IT,
            Marketplace.BE,
            Marketplace.NL,
        ]:
            return Currency.EUR
        if self.value == Marketplace.MX:
            return Currency.MXN
        if self.value == Marketplace.BR:
            return Currency.BRL
        if self.value == Marketplace.CN:
            return Currency.CNY
        if self.value == Marketplace.EG:
            return Currency.EGP
        if self.value == Marketplace.TR:
            return Currency.TRY
        if self.value == Marketplace.SA:
            return Currency.SAR
        if self.value == Marketplace.AE:
            return Currency.AED
        if self.value == Marketplace.IN:
            return Currency.INR
        if self.value == Marketplace.SG:
            return Currency.SGD
        if self.value == Marketplace.AU:
            return Currency.AUD
        raise keble_exceptions.UnhandledScenarioOrCase(
            admin_note={"self.value": self.value},
            alert_admin=True,
            unhandled_case=f"marketplace={self}",
        )

    def to_amazon_marketplace(self) -> "AmazonMarketplace":
        if self.value in list(AmazonMarketplace):
            return AmazonMarketplace[self.value]
        raise keble_exceptions.ServerSideInvalidParams(
            invalid_params=f"marketplace: {self.value}",
            expected="marketplace in list(AmazonMarketplace)",
            but_got=f"marketplace={self.value}",
            admin_note=self.value,
            alert_admin=True,
        )
        # raise ValueError(
        #     f"[Typing] Failed to convert {self.value} marketplace to amazon marketplace"
        # )

    def to_amazon_marketplace_eu(self) -> "AmazonMarketplaceEU":
        return AmazonMarketplaceEU.from_marketplace(self)

    @classmethod
    def from_amazon_marketplace(
        cls, marketplace: Union[None, "AmazonMarketplace", "AmazonMarketplaceEU"]
    ) -> "Marketplace":
        if marketplace is None or marketplace.value not in list(Marketplace):
            raise keble_exceptions.ClientSideInvalidParams(
                invalid_params="marketplace",
                admin_note=marketplace,
                alert_admin=True,
                expected="marketplace is None or marketplace.value not in list(Marketplace)",
                but_got=f"marketplace={marketplace}",
            )
            # raise ValueError(
            #     f"[Typing] Failed to convert amazon marketplace {marketplace} to marketplace enum"
            # )
        return Marketplace[marketplace.value]


AnyMarketplace = (
    Marketplace | UnspecifiedMarketplace
)  # include marketplace, unspecified marketplace (amazon and eu marketplace are subsets of marketplace)

class AmazonProductUrlParsedResult(BaseModel):
    model_config = PydanticModelConfig.default()
    asin: str
    marketplace: "AmazonMarketplace"

class AmazonMarketplace(str, Enum):
    # NA
    US = "US"
    CA = "CA"
    MX = "MX"

    # ASIA
    JP = "JP"

    # EU + UK
    UK = "UK"
    DE = "DE"
    ES = "ES"
    FR = "FR"
    IT = "IT"
    NL = "NL"
    SE = "SE"
    PL = "PL"
    BE = "BE"

    # Unsupported marketplace:

    # IN = "IN"
    # CN = "CN"
    # SG = "SG"

    # AE = "AE"
    # BR = "BR"
    # AU = "AU"
    # TR = "TR"
    # SA = "SA"

    @property
    def currency(self) -> "Currency":
        if self.value == Marketplace.US:
            return Currency.USD
        if self.value == Marketplace.JP:
            return Currency.JPY
        if self.value == Marketplace.UK:
            return Currency.GBP
        if self.value == Marketplace.SE:
            return Currency.SEK
        if self.value == Marketplace.PL:
            return Currency.PLN
        if self.value in [
            Marketplace.DE,
            Marketplace.ES,
            Marketplace.FR,
            Marketplace.IT,
            Marketplace.BE,
            Marketplace.NL,
        ]:
            return Currency.EUR
        if self.value == Marketplace.CA:
            return Currency.CAD
        if self.value == Marketplace.MX:
            return Currency.MXN
        raise keble_exceptions.UnhandledScenarioOrCase(
            admin_note={"self.value": self.value},
            alert_admin=True,
            unhandled_case=f"marketplace={self}",
        )

    def to_marketplace(self) -> Marketplace:
        return Marketplace[self.value]

    def to_amazon_marketplace_eu(self) -> "AmazonMarketplaceEU":
        return AmazonMarketplaceEU.from_marketplace(self)

    # @classmethod
    # def from_marketplace(
    #     cls, marketplace: Union[None, Marketplace, "AmazonMarketplaceEU"]
    # ) -> "AmazonMarketplace":
    #     if marketplace is None or marketplace.value not in list(AmazonMarketplace):
    #         raise keble_exceptions.ServerSideInvalidParams(
    #             invalid_params="marketplace",
    #             admin_note=marketplace,
    #             alert_admin=True,
    #             but_got=f"marketplace={marketplace}",
    #             expected="not(marketplace is None or marketplace.value not in list(AmazonMarketplace))",
    #         )
    #         # raise ValueError(
    #         #     f"[Typing] Failed to convert marketplace {marketplace} to amazon marketplave enum"
    #         # )
    #     return AmazonMarketplace[marketplace.value]

    @classmethod
    def from_marketplace(
        cls,
        marketplace: Union[
            str, Marketplace, "AmazonMarketplaceEU", "AmazonMarketplace", None
        ],
    ) -> "AmazonMarketplace":
        if isinstance(marketplace, AmazonMarketplace):
            return marketplace

        if isinstance(marketplace, (Marketplace, AmazonMarketplaceEU)):
            if marketplace.value not in list(AmazonMarketplace):
                raise keble_exceptions.ServerSideInvalidParams(
                    invalid_params="marketplace",
                    admin_note={"marketplace": str(marketplace)},
                    alert_admin=True,
                    but_got=f"{marketplace}",
                    expected="Valid Marketplace or AmazonMarketplaceEU value in AmazonMarketplace",
                )
            return cls[marketplace.value]

        if isinstance(marketplace, str):
            if marketplace.upper() not in [am.value for am in list(AmazonMarketplace)]:
                raise keble_exceptions.ServerSideInvalidParams(
                    invalid_params="marketplace",
                    admin_note={"marketplace": marketplace},
                    alert_admin=True,
                    but_got=marketplace,
                    expected="Valid AmazonMarketplace string",
                )
            return cls(marketplace.upper())

        raise keble_exceptions.ServerSideInvalidParams(
            invalid_params="marketplace",
            admin_note={
                "marketplace": str(marketplace),
                "type": str(type(marketplace)),
            },
            alert_admin=True,
            but_got=str(marketplace),
            expected="AmazonMarketplace | Marketplace | AmazonMarketplaceEU | str",
        )

    @property
    def is_eu(self):
        eu_values = [e.value for e in list(AmazonMarketplaceEU)]
        return self.value in eu_values

    @property
    def domain(self) -> str:
        domain_mapping = {
            AmazonMarketplace.US.value: "amazon.com",
            AmazonMarketplace.CA.value: "amazon.ca",
            AmazonMarketplace.MX.value: "amazon.com.mx",
            AmazonMarketplace.UK.value: "amazon.co.uk",
            AmazonMarketplace.DE.value: "amazon.de",
            AmazonMarketplace.FR.value: "amazon.fr",
            AmazonMarketplace.IT.value: "amazon.it",
            AmazonMarketplace.ES.value: "amazon.es",
            AmazonMarketplace.NL.value: "amazon.nl",
            AmazonMarketplace.SE.value: "amazon.se",
            AmazonMarketplace.PL.value: "amazon.pl",
            AmazonMarketplace.BE.value: "amazon.com.be",
            AmazonMarketplace.JP.value: "amazon.co.jp",
            "IN": "amazon.in",
            "AU": "amazon.com.au",
            "AE": "amazon.ae",
            "SG": "amazon.sg",
            "TR": "amazon.com.tr",
            "SA": "amazon.sa",
            "BR": "amazon.com.br",
        }
        return domain_mapping[self.value]
    
    @classmethod
    def parse_amazon_product_url(cls, url: str | AnyHttpUrl) -> AmazonProductUrlParsedResult | None:
        raw_url = str(url).strip()
        if raw_url == "":
            return None

        parsed_url = urlsplit(raw_url)
        host = parsed_url.netloc.lower()
        if host == "":
            return None

        host = host.split("@")[-1]
        host = host.split(":")[0]
        while host.startswith("www."):
            host = host[4:]

        domain_to_marketplace: dict[str, AmazonMarketplace] = {
            marketplace.domain: marketplace for marketplace in cls
        }

        marketplace: AmazonMarketplace | None = None
        for domain, candidate_marketplace in sorted(
            domain_to_marketplace.items(),
            key=lambda pair: len(pair[0]),
            reverse=True,
        ):
            if host == domain or host.endswith(f".{domain}"):
                marketplace = candidate_marketplace
                break

        if marketplace is None:
            return None

        path = parsed_url.path
        if path == "":
            return None

        asin: str | None = None
        product_path_patterns = (
            r"/dp/([A-Z0-9]{10})(?:[/?]|$)",
            r"/gp/product/([A-Z0-9]{10})(?:[/?]|$)",
            r"/gp/aw/d/([A-Z0-9]{10})(?:[/?]|$)",
            r"/gp/aw/dp/([A-Z0-9]{10})(?:[/?]|$)",
            r"/gp/dp/([A-Z0-9]{10})(?:[/?]|$)",
            r"/exec/obidos/asin/([A-Z0-9]{10})(?:[/?]|$)",
            r"/o/asin/([A-Z0-9]{10})(?:[/?]|$)",
            r"/product/([A-Z0-9]{10})(?:[/?]|$)",
            r"/product-reviews/([A-Z0-9]{10})(?:[/?]|$)",
            r"/gp/offer-listing/([A-Z0-9]{10})(?:[/?]|$)",
            r"/gp/customer-reviews/([A-Z0-9]{10})(?:[/?]|$)",
        )
        for pattern in product_path_patterns:
            match = re.search(pattern, path, flags=re.IGNORECASE)
            if match is not None:
                asin = match.group(1).upper()
                break

        if asin is None:
            for query_key, query_value in parse_qsl(parsed_url.query):
                if query_key.lower() != "asin":
                    continue
                if re.fullmatch(r"[A-Z0-9]{10}", query_value, flags=re.IGNORECASE):
                    asin = query_value.upper()
                    break

        if asin is None:
            return None

        return AmazonProductUrlParsedResult(asin=asin, marketplace=marketplace)


class AmazonMarketplaceEU(str, Enum):
    DE = AmazonMarketplace.DE.value
    IT = AmazonMarketplace.IT.value
    FR = AmazonMarketplace.FR.value
    ES = AmazonMarketplace.ES.value
    UK = AmazonMarketplace.UK.value
    NL = AmazonMarketplace.NL.value
    SE = AmazonMarketplace.SE.value
    PL = AmazonMarketplace.PL.value
    BE = AmazonMarketplace.BE.value

    @property
    def currency(self) -> Optional["Currency"]:
        if self.value == AmazonMarketplaceEU.UK:
            return Currency.GBP
        if self.value == AmazonMarketplaceEU.SE:
            return Currency.SEK
        if self.value == AmazonMarketplaceEU.PL:
            return Currency.PLN
        if self.value in [
            AmazonMarketplaceEU.DE,
            AmazonMarketplaceEU.ES,
            AmazonMarketplaceEU.FR,
            AmazonMarketplaceEU.IT,
            AmazonMarketplaceEU.BE,
        ]:
            return Currency.EUR
        return None

    def to_marketplace(self) -> Marketplace:
        return Marketplace[self.value]

    def to_amazon_marketplace(self) -> AmazonMarketplace:
        return AmazonMarketplace[self.value]

    @classmethod
    def from_marketplace(
        cls, marketplace: AmazonMarketplace | Marketplace | None
    ) -> "AmazonMarketplaceEU":
        if marketplace is None or marketplace.value not in list(AmazonMarketplaceEU):
            raise keble_exceptions.ServerSideInvalidParams(
                invalid_params="marketplace",
                expected="not(marketplace is None or marketplace.value not in list(AmazonMarketplaceEU))",
                but_got=f"marketplace={marketplace}",
                alert_admin=True,
            )
            # raise ValueError(
            #     f"[Typing] Failed to convert marketplace {marketplace} to amazon marketplave EU enum"
            # )
        return AmazonMarketplaceEU[marketplace.value]

    @classmethod
    def is_eu_marketplace(
        cls,
        marketplace: Union[
            str, AmazonMarketplace, Marketplace, "AmazonMarketplaceEU", None
        ],
    ) -> bool:
        return marketplace in list(AmazonMarketplaceEU)

    @property
    def domain(self) -> str:
        return self.to_amazon_marketplace().domain


class Currency(str, Enum):
    """
    Currencies
    full list: https://en.wikipedia.org/wiki/ISO_4217#Active_codes
    """

    AED = "AED"
    AFN = "AFN"
    ALL = "ALL"
    AMD = "AMD"
    ANG = "ANG"
    AOA = "AOA"
    ARS = "ARS"
    AUD = "AUD"
    AWG = "AWG"
    AZN = "AZN"
    BAM = "BAM"
    BBD = "BBD"
    BDT = "BDT"
    BGN = "BGN"
    BHD = "BHD"
    BIF = "BIF"
    BMD = "BMD"
    BND = "BND"
    BOB = "BOB"
    BRL = "BRL"
    BSD = "BSD"
    BTN = "BTN"
    BWP = "BWP"
    BYN = "BYN"
    BYR = "BYR"
    BZD = "BZD"
    CAD = "CAD"
    CDF = "CDF"
    CHF = "CHF"
    CLP = "CLP"
    CNY = "CNY"
    COP = "COP"
    CRC = "CRC"
    CVE = "CVE"
    CZK = "CZK"
    DJF = "DJF"
    DKK = "DKK"
    DOP = "DOP"
    DZD = "DZD"
    EGP = "EGP"
    ERN = "ERN"
    ETB = "ETB"
    EUR = "EUR"
    FJD = "FJD"
    FKP = "FKP"
    GBP = "GBP"
    GEL = "GEL"
    GHS = "GHS"
    GIP = "GIP"
    GMD = "GMD"
    GNF = "GNF"
    GTQ = "GTQ"
    GYD = "GYD"
    HKD = "HKD"
    HNL = "HNL"
    HRK = "HRK"
    HTG = "HTG"
    HUF = "HUF"
    IDR = "IDR"
    ILS = "ILS"
    INR = "INR"
    IQD = "IQD"
    IRR = "IRR"
    ISK = "ISK"
    JEP = "JEP"
    JMD = "JMD"
    JOD = "JOD"
    JPY = "JPY"
    KES = "KES"
    KGS = "KGS"
    KHR = "KHR"
    KID = "KID"
    KMF = "KMF"
    KRW = "KRW"
    KWD = "KWD"
    KYD = "KYD"
    KZT = "KZT"
    LAK = "LAK"
    LBP = "LBP"
    LKR = "LKR"
    LRD = "LRD"
    LSL = "LSL"
    LTL = "LTL"
    LVL = "LVL"
    LYD = "LYD"
    MAD = "MAD"
    MDL = "MDL"
    MGA = "MGA"
    MKD = "MKD"
    MMK = "MMK"
    MNT = "MNT"
    MOP = "MOP"
    MRU = "MRU"
    MUR = "MUR"
    MVR = "MVR"
    MWK = "MWK"
    MXN = "MXN"
    MYR = "MYR"
    MZN = "MZN"
    NAD = "NAD"
    NGN = "NGN"
    NIO = "NIO"
    NOK = "NOK"
    NPR = "NPR"
    NZD = "NZD"
    OMR = "OMR"
    PAB = "PAB"
    PEN = "PEN"
    PGK = "PGK"
    PHP = "PHP"
    PKR = "PKR"
    PLN = "PLN"
    PYG = "PYG"
    QAR = "QAR"
    RON = "RON"
    RSD = "RSD"
    RUB = "RUB"
    RWF = "RWF"
    SAR = "SAR"
    SBD = "SBD"
    SCR = "SCR"
    SDG = "SDG"
    SEK = "SEK"
    SGD = "SGD"
    SHP = "SHP"
    SLL = "SLL"
    SOS = "SOS"
    SRD = "SRD"
    SSP = "SSP"
    STD = "STD"
    SYP = "SYP"
    SZL = "SZL"
    THB = "THB"
    TJS = "TJS"
    TMT = "TMT"
    TND = "TND"
    TOP = "TOP"
    TRY = "TRY"
    TTD = "TTD"
    TWD = "TWD"
    TZS = "TZS"
    UAH = "UAH"
    UGX = "UGX"
    USD = "USD"
    UYU = "UYU"
    UZS = "UZS"
    VEF = "VEF"
    VES = "VES"
    VND = "VND"
    VUV = "VUV"
    WST = "WST"
    XAF = "XAF"
    XCD = "XCD"
    XOF = "XOF"
    XPF = "XPF"
    XXX = "XXX"  # 未知货币类型
    YER = "YER"
    ZAR = "ZAR"
    ZMW = "ZMW"

    @property
    def symbol(self):
        _map = {
            "USD": "$",  # US Dollar
            "CAD": "C$",  # Canadian Dollar
            "JPY": "¥",  # Japanese Yen
            "GBP": "£",  # British Pound
            "EUR": "€",  # Euro
            "SEK": "kr",  # Swedish Krona
            "PLN": "zł",  # Polish Zloty
            "MXN": "$",  # mexico peso
        }
        if self.value in _map:
            return _map[self.value]
        return self.value  # by default keeping the same

    @property
    def has_decimal(self) -> bool:
        """Returns True if the currency supports decimal places, otherwise False.
        Function generated by CHATGPT:
        prompts:
            which currencies like JPY, has no floating in decimal

            give me a function
            def has_decial(self) -> bool: ...
        """
        NON_DECIMAL_CURRENCIES = {
            "JPY",
            "KRW",
            "VND",
            "XAF",
            "XOF",
            "CLP",
            "PYG",
            "RWF",
            "BIF",
            "DJF",
            "GNF",
            "KMF",
            "UGX",
            "MGA",
            "VUV",
        }
        return self.value not in NON_DECIMAL_CURRENCIES


class ExchangeRateInUsd(BaseModel):
    # an object representing any currency that can exchange to USD
    # with a list of this object, we can convert different
    # currencies into 1 currency

    model_config = DEFAULT_MODEL_CONFIG
    currency: Currency
    exchange_rate: StdMoneyAmount  # 1 currency can have how much USD. For example, if currency is CNY, this value can be approx 700 (for 7 cny)

    # def __init__(self, *, exchange_rate: StdMoneyAmount, currency: Currency):
    #     super().__init__(exchange_rate=exchange_rate, currency=currency)


class Money(BaseModel):
    amount: StdMoneyAmount
    currency: Currency

    def __init__(
        self,
        *,
        amount: Optional[StdMoneyAmount] = None,
        currency: Currency,
        float_money: Optional[float] = None,
    ):
        if amount is not None:
            super().__init__(amount=amount, currency=currency)
        elif float_money is not None:
            super().__init__(
                amount=StdMoneyAmount.from_true_value(float_money=float_money),
                currency=currency,
            )
        else:
            raise keble_exceptions.ServerSideMissingParams(
                alert_admin=True,
                missing_params="amount or float_money",
                admin_note=f"amount={amount}, float_money={float_money}",
            )
            # raise ValueError(
            #     "[Typing] You need to provide either amount or float_money to create an instance of Money"
            # )

    def exchange_to(
        self, exchange_rates: List[ExchangeRateInUsd], currency: Currency
    ) -> "Money":
        """Ensure money are in the target currency"""
        if self.currency == currency:
            return Money(**self.model_dump())  # make a clone
        usd_amount = None
        for er in exchange_rates:
            if er.currency == self.currency:
                usd_amount = self.amount / er.exchange_rate
        # assert usd_amount is not None, (
        #     f"[Typing] Failed to exchange money {self} to currency {currency}, due to {self.currency} to USD exchange rate not found."
        # )
        keble_exceptions.raise_if_not(
            usd_amount is not None,
            keble_exceptions.ServerSideMissingParams(
                missing_params=f"USD exchange rate for {self.currency}",
                admin_note={"self.currency": self.currency, "currency": currency},
                alert_admin=True,
            ),
        )

        for er in exchange_rates:
            if er.currency == currency:
                return Money(amount=er.exchange_rate * usd_amount, currency=currency)

        raise keble_exceptions.ServerSideMissingParams(
            missing_params=f"USD exchange rate for {currency}",
            admin_note={"self.currency": self.currency, "currency": currency},
            alert_admin=True,
        )
        # raise AssertionError(
        #     f"[Typing] Failed to exchange money {self} to currency {currency}, due to {currency} to USD exchange rate not found."
        # )

    def _check_currency_match(self, other):
        """Helper method to check if the currencies match"""
        if self.currency != other.currency:
            raise keble_exceptions.ServerSideInvalidParams(
                invalid_params="two currency values",
                admin_note=f"self.currency={self.currency}, other.currency={other.currency}",
                alert_admin=True,
                but_got=f"self.currency={self.currency}, other.currency={other.currency}",
                expected="self.currency == other.currency",
            )
            # raise ValueError(
            #     f"[Typing] Failed to do operation due to currency unmatched. Left currency {self.currency} and right currency {other.currency}"
            # )

    def __add__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return Money(amount=self.amount + other.amount, currency=self.currency)
        raise NotImplementedError

    def __sub__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return Money(amount=self.amount - other.amount, currency=self.currency)
        raise NotImplementedError

    def __mul__(self, other):
        if isinstance(other, int):
            return Money(amount=self.amount * other, currency=self.currency)
        raise NotImplementedError

    def __truediv__(self, other):
        if isinstance(other, int):
            return Money(amount=self.amount / other, currency=self.currency)
        raise NotImplementedError

    def __mod__(self, other):
        if isinstance(other, (int, float)):
            return self.true_value % other
        raise NotImplementedError

    def __floordiv__(self, other):
        if isinstance(other, int):
            return Money(amount=self.amount // other, currency=self.currency)
        raise NotImplementedError

    def __pow__(self, other):
        if isinstance(other, int):
            return Money(amount=self.amount**other, currency=self.currency)
        raise NotImplementedError

    def __neg__(self):
        return Money(amount=-self.amount, currency=self.currency)

    def __abs__(self):
        return Money(amount=abs(self.amount), currency=self.currency)

    def __lt__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return self.amount < other.amount
        elif isinstance(other, int):
            return self.amount < other
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return self.amount <= other.amount
        elif isinstance(other, int):
            return self.amount <= other
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return self.amount > other.amount
        elif isinstance(other, int):
            return self.amount > other
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return self.amount >= other.amount
        elif isinstance(other, int):
            return self.amount >= other
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return self.amount == other.amount
        elif isinstance(other, int):
            return self.amount == other
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Money):
            self._check_currency_match(other)
            return self.amount != other.amount
        elif isinstance(other, int):
            return self.amount != other
        return NotImplemented

    def __str__(self):
        return f"{self.true_value} {self.currency.value}"

    def __repr__(self):
        return f"Money(amount={self.amount}, currency={self.currency})"

    @property
    def true_value(self) -> float | int:
        # 100% == 10000 -> 1, 10% == 1000 -> 0.1
        if self.currency.has_decimal:
            return round(self.amount.true_value, 2)
        return int(round(self.amount.true_value, 0))

    @classmethod
    def from_true_value(cls, *, float_money: float, currency: Currency):
        return cls(
            amount=StdMoneyAmount.from_true_value(float_money=float_money),
            currency=currency,
        )

    @property
    def string(self) -> str:
        return f"{self.currency.symbol if self.currency.symbol is not None else self.currency.value}{self.true_value}"


class MoneyRange(BaseModel):
    model_config = DEFAULT_MODEL_CONFIG

    lower: Optional[Money]  # gte
    upper: Optional[Money]  # lte
    currency: Currency

    @model_validator(mode="after")
    def validate_self(self):
        if self.lower is not None and self.lower.currency != self.currency:
            raise keble_exceptions.ServerSideInvalidParams(
                invalid_params="lower.currency",
                admin_note={
                    "lower.currency": self.lower.currency,
                    "currency": self.currency,
                },
                alert_admin=True,
                but_got=f"lower.currency={self.lower.currency}",
                expected=f"lower.currency == {self.currency}",
            )

        if self.upper is not None and self.upper.currency != self.currency:
            raise keble_exceptions.ServerSideInvalidParams(
                invalid_params="upper.currency",
                admin_note={
                    "upper.currency": self.upper.currency,
                    "currency": self.currency,
                },
                alert_admin=True,
                but_got=f"upper.currency={self.upper.currency}",
                expected=f"upper.currency == {self.currency}",
            )

        if self.lower is not None and self.upper is not None:
            if self.lower.currency != self.upper.currency:
                raise keble_exceptions.ServerSideInvalidParams(
                    invalid_params="lower.currency and upper.currency",
                    admin_note={
                        "lower.currency": self.lower.currency,
                        "upper.currency": self.upper.currency,
                    },
                    alert_admin=True,
                    but_got=f"lower.currency={self.lower.currency} upper.currency={self.upper.currency}",
                    expected="lower.currency == upper.currency",
                )

            if self.lower > self.upper:
                raise keble_exceptions.ServerSideInvalidParams(
                    invalid_params="lower and upper",
                    admin_note={"lower": str(self.lower), "upper": str(self.upper)},
                    alert_admin=True,
                    but_got="lower > upper",
                    expected="lower <= upper",
                )
        return self

    @classmethod
    def build(
        cls,
        *,
        lower: Optional[Money],
        upper: Optional[Money],
        currency: Currency,
    ):
        return MoneyRange(
            lower=lower,
            upper=upper,
            currency=currency,
        )

    @property
    def string(self):
        prefix = (
            self.currency.symbol
            if self.currency.symbol is not None
            else self.currency.value
        )
        if (
            self.lower is not None
            and self.upper is not None
            and self.lower != self.upper
        ):
            return f"{self.lower.string} ~ {self.upper.string}"
        elif self.lower is not None or self.upper is not None:
            if self.lower is not None:
                return self.lower.string
            assert self.upper is not None
            return self.upper.string
        return f"{prefix}-"

    def __str__(self):
        if self.lower is None and self.upper is None:
            return "-"
        elif self.lower is not None and self.upper is not None:
            return f"{str(self.lower)} ~ {str(self.upper)}"
        elif self.lower is not None:
            return f">= {str(self.lower)}"
        assert self.upper is not None
        return f"<= {str(self.upper)}"


class PercentageRange(BaseModel):
    model_config = DEFAULT_MODEL_CONFIG

    lower: Optional[StdPercentage]  # gte
    upper: Optional[StdPercentage]  # lte

    @model_validator(mode="after")
    def validate_self(self):
        if (
            self.lower is not None
            and self.upper is not None
            and self.lower > self.upper
        ):
            raise keble_exceptions.ServerSideInvalidParams(
                invalid_params="lower and upper",
                admin_note={"lower": str(self.lower), "upper": str(self.upper)},
                alert_admin=True,
                but_got="lower > upper",
                expected="lower <= upper",
            )
        return self

    @classmethod
    def build(cls, *, lower: Optional[StdPercentage], upper: Optional[StdPercentage]):
        return PercentageRange(
            lower=lower,
            upper=upper,
        )

    @property
    def string(self):
        if (
            self.lower is not None
            and self.upper is not None
            and self.lower != self.upper
        ):
            return f"{self.lower.string} ~ {self.upper.string}"
        elif self.lower is not None or self.upper is not None:
            if self.lower is not None:
                return self.lower.string
            assert self.upper is not None
            return self.upper.string
        return "-%"

    def __str__(self):
        if self.lower is None and self.upper is None:
            return "-"
        elif self.lower is not None and self.upper is not None:
            return f"{str(self.lower)} ~ {str(self.upper)}"
        elif self.lower is not None:
            return f">= {str(self.lower)}"
        assert self.upper is not None
        return f"<= {str(self.upper)}"

    def __repr__(self):
        return f"PercentageRange(lower={self.lower}, upper={self.upper})"


class TimestampValidator:
    @classmethod
    def validate(cls, v: Any) -> Any:
        if isinstance(v, (int, float)):
            return datetime.fromtimestamp(v, tz=timezone.utc)
        return v

    @classmethod
    def serialize(cls, v: Any, info: Any) -> float:
        if isinstance(v, datetime):
            return v.timestamp()
        else:
            assert isinstance(v, float)
            return v

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_before_validator_function(
            cls.validate,
            core_schema.datetime_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.serialize, info_arg=True
            ),
        )


Timestamp = Annotated[datetime, TimestampValidator()]


class QdrantSchemaBase(BaseModel):
    model_config = DEFAULT_MODEL_CONFIG
    created_ts: Timestamp = Field(default_factory=lambda: datetime.now(timezone.utc))
    updated_ts: Timestamp = Field(default_factory=lambda: datetime.now(timezone.utc))


class StrapiSchemaBase(BaseModel):
    model_config = DEFAULT_MODEL_CONFIG
    id: int  # auto increment
    # strapi fields
    document_id: str
    created_at: Optional[datetime]
    updated_at: Optional[datetime]
    published_at: Optional[datetime]
    locale: str
    created_by_id: Optional[int]
    updated_by_id: Optional[int]


class SchemaBase(BaseModel):
    model_config = DEFAULT_MODEL_CONFIG
    created: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    updated: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))

    @field_validator("created", "updated", mode="before")
    @classmethod
    def _ensure_utc(cls, v):
        # If no value provided, let the default_factory run
        if v is None:
            return v

        # parse string if needed (pydantic will already do that by default)
        # at this point v should be a datetime
        if isinstance(v, datetime):
            # if tzinfo is missing, assign UTC
            if v.tzinfo is None:
                return v.replace(tzinfo=timezone.utc)
            # otherwise leave it alone
            return v

        return v  # let pydantic handle invalid types later


class MongoObjectBase(SchemaBase):
    model_config = DEFAULT_MODEL_CONFIG
    id: ObjectId = Field(
        ...,
        validation_alias=AliasChoices("_id", "id"),
        pattern=r"^[0-9a-fA-F]{24}$",
    )


class Status(str, Enum):
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"


class SharingScope(str, Enum):
    PRIVATE = "PRIVATE"
    PUBLIC = "PUBLIC"


class OwnerType(str, Enum):
    USER = "USER"
    ORG = "ORG"
    ROOT = "ROOT"



from decimal import ROUND_DOWN, Decimal, getcontext
from typing import Optional, TypeVar, Union

import keble_exceptions
from pydantic import BaseModel

# Set precision for decimal calculations
getcontext().prec = 28

T = TypeVar("T", bound="StdMoneyAmount")
P = TypeVar("P", bound="StdPercentage")

Number = Union[int, float, Decimal]


def _to_decimal(value: Number) -> Decimal:
    if isinstance(value, Decimal):
        return value
    if isinstance(value, float):
        return Decimal(str(value))
    return Decimal(value)


def _decimal_fmod(dividend: Decimal, divisor: Decimal) -> Decimal:
    if divisor == 0:
        raise ZeroDivisionError("Modulo by zero")

    quotient = (dividend / divisor).to_integral_value(rounding=ROUND_DOWN)
    return dividend - (quotient * divisor)


class StdMoneyAmount(BaseModel):
    value: int

    def __init__(self, value: int):
        super().__init__(value=value)

    def __add__(self, other):
        if isinstance(other, StdMoneyAmount):
            return StdMoneyAmount(value=self.value + other.value)
        elif isinstance(other, (int, float, Decimal)):
            return StdMoneyAmount(value=self.value + int(_to_decimal(other)))
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, StdMoneyAmount):
            return StdMoneyAmount(value=self.value - other.value)
        elif isinstance(other, (int, float, Decimal)):
            return StdMoneyAmount(value=self.value - int(_to_decimal(other)))
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):
            result = _to_decimal(self.value) * _to_decimal(other)
            return StdMoneyAmount(value=int(result))
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, StdMoneyAmount):
            # Dividing two StdMoneyAmount values gives their ratio
            if other.value == 0:
                raise ZeroDivisionError("Division by zero")
            return float(_to_decimal(self.value) / _to_decimal(other.value))
        elif isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            if other_decimal == 0:
                raise ZeroDivisionError("Division by zero")
            result = _to_decimal(self.value) / other_decimal
            return StdMoneyAmount(value=int(result))
        return NotImplemented

    def __mod__(self, other):
        if isinstance(other, (int, float, Decimal)):
            divisor = _to_decimal(other)
            result = _decimal_fmod(_to_decimal(self.value) / Decimal("100"), divisor)
            return float(result)
        return NotImplemented

    def __floordiv__(self, other):
        if isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            if other_decimal == 0:
                raise ZeroDivisionError("Division by zero")
            result = _to_decimal(self.value) // other_decimal
            return StdMoneyAmount(value=int(result))
        return NotImplemented

    def __pow__(self, other):
        if isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            result = _to_decimal(self.value) ** other_decimal
            return StdMoneyAmount(value=int(result))
        return NotImplemented

    def __neg__(self):
        return StdMoneyAmount(value=-self.value)

    def __abs__(self):
        return StdMoneyAmount(value=abs(self.value))

    def __lt__(self, other):
        if isinstance(other, StdMoneyAmount):
            return self.value < other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value < int(other)
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, StdMoneyAmount):
            return self.value <= other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value <= int(other)
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, StdMoneyAmount):
            return self.value > other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value > int(other)
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, StdMoneyAmount):
            return self.value >= other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value >= int(other)
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, StdMoneyAmount):
            return self.value == other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value == int(other)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, StdMoneyAmount):
            return self.value != other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value != int(other)
        return NotImplemented

    def __str__(self):
        return f"{self.value}"

    def __repr__(self):
        return f"MoneyAmount(value={self.value})"

    @property
    def true_value(self) -> float:
        # 10000 -> 100 USD, 10077 -> 100.77 USD
        # Return float for compatibility with existing tests
        return float(Decimal(self.value) / Decimal("100"))

    @classmethod
    def from_true_value(cls: type[T], *, float_money: float) -> T:
        decimal_money = _to_decimal(float_money)
        return cls(value=int(decimal_money * Decimal("100")))


class StdPercentage(BaseModel):
    value: int  # Represent 100% as 10000

    def __init__(self, value: int):
        super().__init__(value=value)

    def __add__(self, other):
        if isinstance(other, StdPercentage):
            return StdPercentage(value=self.value + other.value)
        elif isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            return StdPercentage(
                value=self.value + int(other_decimal * Decimal("10000"))
            )
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, StdPercentage):
            return StdPercentage(value=self.value - other.value)
        elif isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            return StdPercentage(
                value=self.value - int(other_decimal * Decimal("10000"))
            )
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            result = _to_decimal(self.value) * other_decimal
            return StdPercentage(value=int(result))
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            if other_decimal == 0:
                raise ZeroDivisionError("Division by zero")
            result = _to_decimal(self.value) / other_decimal
            return StdPercentage(value=int(result))
        return NotImplemented

    def __mod__(self, other):
        if isinstance(other, (int, float, Decimal)):
            divisor_bps = _to_decimal(other)
            result_percent = _decimal_fmod(
                _to_decimal(self.value) / Decimal("100"),
                divisor_bps / Decimal("100"),
            )
            return float(result_percent)
        return NotImplemented

    def __floordiv__(self, other):
        if isinstance(other, (int, float, Decimal)):
            other_decimal = _to_decimal(other)
            if other_decimal == 0:
                raise ZeroDivisionError("Division by zero")
            result = _to_decimal(self.value) // other_decimal
            return StdPercentage(value=int(result))
        return NotImplemented

    def __pow__(self, other):
        if isinstance(other, (int, float, Decimal)):
            other_decimal = (
                Decimal(str(other)) if not isinstance(other, Decimal) else other
            )
            result = Decimal(self.value) ** other_decimal
            return StdPercentage(value=int(result))
        return NotImplemented

    def __neg__(self):
        return StdPercentage(value=-self.value)

    def __abs__(self):
        return StdPercentage(value=abs(self.value))

    def __lt__(self, other):
        if isinstance(other, StdPercentage):
            return self.value < other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value < int(_to_decimal(other) * Decimal("10000"))
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, StdPercentage):
            return self.value <= other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value <= int(_to_decimal(other) * Decimal("10000"))
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, StdPercentage):
            return self.value > other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value > int(_to_decimal(other) * Decimal("10000"))
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, StdPercentage):
            return self.value >= other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value >= int(_to_decimal(other) * Decimal("10000"))
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, StdPercentage):
            return self.value == other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value == int(_to_decimal(other) * Decimal("10000"))
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, StdPercentage):
            return self.value != other.value
        elif isinstance(other, (int, float, Decimal)):
            return self.value != int(_to_decimal(other) * Decimal("10000"))
        return NotImplemented

    def __str__(self):
        decimal_value = (_to_decimal(self.value) / Decimal("100")).quantize(
            Decimal("0.01")
        )
        return f"{decimal_value}%"

    def __repr__(self):
        return f"Percentage(value={self.value})"

    @property
    def true_value(self) -> float:
        # 100% == 10000 integer,
        # or 1 in float == 10000 integer,
        # 50% == 5000 integer
        # or 0.5 in float == 5000 integer
        # Return float for compatibility with existing tests
        return float(_to_decimal(self.value) / Decimal("10000"))

    @classmethod
    def from_true_value(
        cls: type[P],
        *,
        float_percent: Optional[float] = None,
        string_percent: Optional[str] = None,
    ) -> P:
        if float_percent is not None:
            decimal_percent = _to_decimal(float_percent)
            return cls(value=int(decimal_percent * Decimal("10000")))
        elif string_percent is not None:
            decimal_percent = Decimal(string_percent.strip().rstrip("%"))
            return cls(value=int(decimal_percent * Decimal("100")))
        else:
            raise keble_exceptions.ServerSideMissingParams(
                alert_admin=True, missing_params="float_percent or string_percent"
            )

    @property
    def string(self) -> str:
        decimal_value = (_to_decimal(self.value) / Decimal("100")).quantize(
            Decimal("0.01")
        )
        _str = format(decimal_value, "f")
        if "." in _str:
            _str = _str.rstrip("0").rstrip(".")
        return f"{_str}%"

    @property
    def four_digits_string(self) -> str:
        """non stripped version of string"""
        decimal_value = (_to_decimal(self.value) / Decimal("100")).quantize(
            Decimal("0.01")
        )
        return f"{decimal_value}%"



from .cloud_storage import CloudStorageBase, CloudStorageObjectType, CloudStorageType
from .config import PydanticModelConfig
from .id import MaybeUuid, ObjectId, Uuid, is_optional, to_binary_uuid
from .id import _ObjectId as BsonObjectId
from .schemas import is_http_url
from .standard_number import Number, StdMoneyAmount, StdPercentage
from .typing import (
    AmazonMarketplace,
    AmazonMarketplaceEU,
    AnyMarketplace,
    Currency,
    ExchangeRateInUsd,
    Language,
    Marketplace,
    Money,
    MoneyRange,
    MongoObjectBase,
    NaturalLanguage,
    OneOrMoreVectors,
    OwnerType,
    PercentageRange,
    PromptRole,
    QdrantSchemaBase,
    SchemaBase,
    SharingScope,
    Status,
    StrapiSchemaBase,
    Timestamp,
    UnspecifiedMarketplace,
    Vector,
    is_vectors_list,
)

__all__ = [
    "CloudStorageBase",
    "CloudStorageObjectType",
    "CloudStorageType",
    "PydanticModelConfig",
    "ObjectId",
    "Uuid",
    "MaybeUuid",
    "is_optional",
    "to_binary_uuid",
    "is_http_url",
    "Number",
    "StdMoneyAmount",
    "StdPercentage",
    "AmazonMarketplace",
    "AmazonMarketplaceEU",
    "Currency",
    "ExchangeRateInUsd",
    "Language",
    "Marketplace",
    "Money",
    "MoneyRange",
    "MongoObjectBase",
    "NaturalLanguage",
    "OneOrMoreVectors",
    "OwnerType",
    "PercentageRange",
    "PromptRole",
    "QdrantSchemaBase",
    "SchemaBase",
    "SharingScope",
    "Status",
    "StrapiSchemaBase",
    "Timestamp",
    "Vector",
    "is_vectors_list",
    "BsonObjectId",
    "AnyMarketplace",
    "UnspecifiedMarketplace",
]
